# pc/audio/audio_processor.py
import pyaudio
import numpy as np
import queue
import logging
from dataclasses import dataclass
from typing import Optional
import time
import threading
from scipy import signal

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

@dataclass
class AudioState:
    """PC側の音声状態を管理するデータクラス"""
    mic_enabled: bool = True
    speaker_enabled: bool = True
    volume: int = 50
    input_device_index: Optional[int] = None
    output_device_index: Optional[int] = None
    last_packet_time: float = 0.0
    packet_count: int = 0

class PCAudioProcessor:
    def __init__(self):
        self.CHUNK = 1024
        self.FORMAT = pyaudio.paFloat32
        self.CHANNELS = 1
        self.input_RATE = 44100
        self.output_RATE = 48000
        
        # デバイスインデックスを直接指定
        self.input_device_index = 1  # PCマイクのデバイスインデックス
        self.output_device_index = 8  # PCスピーカーのデバイスインデックス
        
        self.p = None
        self.input_stream = None
        self.output_stream = None
        
        self.state = AudioState(
            input_device_index=self.input_device_index,
            output_device_index=self.output_device_index
        )
        
        self.running = False
        self.input_queue = queue.Queue(maxsize=10)
        self.output_queue = queue.Queue(maxsize=10)

    def initialize(self):
        """音声システムの初期化"""
        try:
            self.p = pyaudio.PyAudio()
            self._start_streams()
            self.running = True
            
            threading.Thread(target=self._input_processor, daemon=True).start()
            threading.Thread(target=self._output_processor, daemon=True).start()
            
            logger.info("PC audio processor initialized")
        except Exception as e:
            logger.error(f"Failed to initialize PC audio: {e}")
            self.cleanup()
            raise

    def _start_streams(self):
        """オーディオストリームの開始"""
        try:
            self.input_stream = self.p.open(
                format=self.FORMAT,
                channels=self.CHANNELS,
                rate=self.input_RATE,
                input=True,
                input_device_index=self.input_device_index,
                frames_per_buffer=self.CHUNK
            )
            
            self.output_stream = self.p.open(
                format=self.FORMAT,
                channels=self.CHANNELS,
                rate=self.output_RATE,
                output=True,
                output_device_index=self.output_device_index,
                frames_per_buffer=self.CHUNK
            )
        except Exception as e:
            logger.error(f"Failed to start streams: {e}")
            raise

    def _input_processor(self):
        """マイク入力の処理"""
        while self.running:
            if self.state.mic_enabled:
                try:
                    data = self.input_stream.read(self.CHUNK, exception_on_overflow=False)
                    if not self.input_queue.full():
                        self.input_queue.put_nowait(data)
                except queue.Full:
                    continue
                except Exception as e:
                    logger.error(f"Input error: {e}")

    def _output_processor(self):
        """スピーカー出力の処理"""
        while self.running:
            if self.state.speaker_enabled:
                try:
                    data = self.output_queue.get(timeout=0.1)
                    audio_array = np.frombuffer(data, dtype=np.float32)
                    volume_adjusted = (audio_array * (self.state.volume / 100)).astype(np.float32)
                    self.output_stream.write(volume_adjusted.tobytes())
                except queue.Empty:
                    continue
                except Exception as e:
                    logger.error(f"Output error: {e}")

    def process_incoming_audio(self, audio_data):
        """受信した音声データの処理"""
        if self.state.speaker_enabled:
            try:
                if not self.output_queue.full():
                    self.output_queue.put_nowait(audio_data)
            except queue.Full:
                pass

    def get_outgoing_audio(self):
        """送信する音声データの取得"""
        try:
            return self.input_queue.get_nowait()
        except queue.Empty:
            return None

    def set_volume(self, volume: int):
        """音量の設定"""
        self.state.volume = max(0, min(100, volume))
        logger.info(f"PC volume set to {self.state.volume}")

    def set_mic_enabled(self, enabled: bool):
        """マイクのON/OFF切り替え"""
        self.state.mic_enabled = enabled
        logger.info(f"PC microphone {'enabled' if enabled else 'disabled'}")

    def set_speaker_enabled(self, enabled: bool):
        """スピーカーのON/OFF切り替え"""
        self.state.speaker_enabled = enabled
        logger.info(f"PC speaker {'enabled' if enabled else 'disabled'}")

    def cleanup(self):
        """リソースの解放"""
        self.running = False
        if self.input_stream:
            self.input_stream.stop_stream()
            self.input_stream.close()
        if self.output_stream:
            self.output_stream.stop_stream()
            self.output_stream.close()
        if self.p:
            self.p.terminate()
        logger.info("PC audio processor cleaned up")