# pc/audio/audio_server.py
from flask_socketio import emit
import logging
from .audio_processor import PCAudioProcessor
import threading
import time

logger = logging.getLogger(__name__)

class AudioServer:
    def __init__(self, socketio):
        self.socketio = socketio
        self.audio_processor = PCAudioProcessor()
        self.setup_socket_events()
        self.running = False

    def setup_socket_events(self):
        @self.socketio.on('connect')
        def handle_connect():
            logger.info("Client connected")
            self._send_status_update()

        @self.socketio.on('disconnect')
        def handle_disconnect():
            logger.info("Client disconnected")

        @self.socketio.on('pc_mic_toggle')
        def handle_pc_mic_toggle(data):
            self.audio_processor.set_mic_enabled(data['enabled'])
            self._send_status_update()

        @self.socketio.on('pc_speaker_toggle')
        def handle_pc_speaker_toggle(data):
            self.audio_processor.set_speaker_enabled(data['enabled'])
            self._send_status_update()

        @self.socketio.on('pc_volume_change')
        def handle_pc_volume_change(data):
            self.audio_processor.set_volume(data['volume'])
            self._send_status_update()

        @self.socketio.on('audio_data')
        def handle_audio_data(data):
            try:
                audio_data = bytes.fromhex(data['audio'])
                self.audio_processor.process_incoming_audio(audio_data)
            except Exception as e:
                logger.error(f"Error processing audio data: {e}")
                emit('error_message', {'message': 'Audio processing error'})

    def _send_status_update(self):
        try:
            state = self.audio_processor.state
            self.socketio.emit('pc_status_update', {
                'mic_enabled': state.mic_enabled,
                'speaker_enabled': state.speaker_enabled,
                'volume': state.volume
            })
        except Exception as e:
            logger.error(f"Failed to send status update: {e}")

    def _send_audio_loop(self):
        while self.running:
            try:
                if self.audio_processor.state.mic_enabled:
                    audio_data = self.audio_processor.get_outgoing_audio()
                    if audio_data:
                        self.socketio.emit('audio_data', {'audio': audio_data.hex()})
            except Exception as e:
                logger.error(f"Error sending audio data: {e}")
            time.sleep(0.02)

    def start(self):
        try:
            self.audio_processor.initialize()
            self.running = True
            threading.Thread(target=self._send_audio_loop, daemon=True).start()
        except Exception as e:
            logger.error(f"Failed to start audio server: {e}")
            self.cleanup()
            raise

    def cleanup(self):
        self.running = False
        if self.audio_processor:
            self.audio_processor.cleanup()