import subprocess
import os
import signal
import atexit
import paramiko
import sys
import time
import socket
import requests
import logging
import logging.handlers
import io
import locale
import threading
from threading import Thread, Event
from flask import Flask, render_template, request, jsonify
from flask_cors import CORS
from flask_socketio import SocketIO

detection_state = {
    'camera1_detected': False,
    'camera2_detected': False
}

def setup_logging():
    """
    UTF-8エンコーディングを考慮した詳細なロギング設定
    """
    # システム全体のエンコーディングを設定
    sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')
    sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding='utf-8')

    # ログディレクトリの作成
    log_dir = 'logs'
    os.makedirs(log_dir, exist_ok=True)
    
    # ログファイルのパス設定
    log_file = os.path.join(log_dir, 'run_apps.log')
    
    # ロガーの取得
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    
    # 既存のハンドラをクリア
    logger.handlers.clear()
    
    # フォーマッターの作成（明示的なUTF-8エンコーディング）
    formatter = logging.Formatter(
        '%(asctime)s - %(levelname)s - [%(filename)s:%(lineno)d] - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    
    # ファイルハンドラ（UTF-8エンコーディング、BOMあり）
    file_handler = logging.FileHandler(
        log_file, 
        encoding='utf-8-sig'  # BOMありのUTF-8
    )
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(formatter)
    
    # コンソールハンドラ
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(formatter)
    
    # ハンドラの追加
    logger.addHandler(file_handler)
    logger.addHandler(console_handler)
    
    # ロケール設定
    locale.setlocale(locale.LC_ALL, '')  # システムのデフォルトロケールを使用
    
    # ロギング開始のメッセージ
    logger.info("=" * 50)
    logger.info("アプリケーションのロギングを初期化")
    logger.info(f"ログファイル: {log_file}")
    logger.info(f"システムエンコーディング: {sys.getdefaultencoding()}")
    logger.info(f"ファイルシステムエンコーディング: {sys.getfilesystemencoding()}")
    logger.info("=" * 50)

    return logger
logger = setup_logging()

# ロギング設定
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    handlers=[
        logging.FileHandler('run_apps.log'),
        logging.StreamHandler()
    ]
)
logger = logging.getLogger(__name__)

# Flask アプリケーションの設定
app = Flask(__name__, template_folder=os.path.join(os.path.dirname(__file__), "templates"))
CORS(app)
socketio = SocketIO(app, cors_allowed_origins="*")

# パス設定

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))  # run_apps.pyの場所
VENV_PYTHON = os.path.join(SCRIPT_DIR, "venv", "Scripts", "python.exe")
WINDOWS_MAIN_PATH = os.path.join(SCRIPT_DIR, "streamer")
WINDOWS_MAIN_SCRIPT = "main.py"

def get_main_script_path():
    return os.path.join(WINDOWS_MAIN_PATH, WINDOWS_MAIN_SCRIPT)

# Raspberry Pi の接続情報
RASPI_HOST = "172.25.16.204"
RASPI_PORT = 22
RASPI_USER = "pi"
RASPI_PASSWORD = "toyogipro"
RASPI_CAMERA_SCRIPT = "/home/pi/test/myproject/camera.py"
RASPI_SERIAL_SCRIPT = "/home/pi/test/myproject/raspi/serial_angle.py"
RASPI_ANGLE_URL = f"http://{RASPI_HOST}:5000/update_angles"

# Flask用
FLASK_PORT = 9000

# グローバル変数
processes = []
ssh_client = None
raspi_pid = None
shutdown_event = Event()
cleanup_done = False

@socketio.on('connect')
def handle_connect():
    logger.info('Client connected')
    # 接続時に現在の状態を送信
    socketio.emit('detection_update', detection_state)

@socketio.on('disconnect')
def handle_disconnect():
    logger.info('Client disconnected')
    
@app.route("/")
def home():
    return render_template("home.html")

@app.route('/detection_status', methods=['GET', 'POST'])
def detection_status():
    global detection_state
    try:
        if request.method == 'POST':
            # main.pyからの更新を受け取る
            new_state = request.get_json()
            logger.info(f"Received detection update: {new_state}")
            detection_state.update(new_state)
            # WebSocket経由でクライアントに通知
            socketio.emit('detection_update', detection_state)
            logger.info(f"Emitted new state: {detection_state}")
            return jsonify({'status': 'success', 'data': detection_state})
        else:
            return jsonify(detection_state)
    except Exception as e:
        logger.error(f"Detection status error: {e}")
        return jsonify({'error': str(e)}), 500

@app.route('/update_angle', methods=['POST'])
def update_angle():
    try:
        data = request.get_json()
        servo1 = data.get("servo1", 25)  # デフォルト値は初期値
        servo2 = data.get("servo2", 17)
        servo3 = data.get("servo3", 137)
        servo4 = data.get("servo4", 120)
        
        try:
            response = requests.post(
                RASPI_ANGLE_URL,
                json={
                    "servo1": servo1,
                    "servo2": servo2,
                    "servo3": servo3,
                    "servo4": servo4
                },
                timeout=5
            )
            if response.status_code == 200:
                raspi_response = response.json()
                return jsonify({
                    "status": "success",
                    "message": "角度データ受信・転送完了",
                    "servo1": servo1,
                    "servo2": servo2,
                    "servo3": servo3,
                    "servo4": servo4,
                    "raspi_response": raspi_response
                }), 200
            else:
                return jsonify({
                    "status": "error",
                    "message": f"Raspberry Piへの送信エラー: {response.status_code}"
                }), 500
            
        except requests.RequestException as e:
            return jsonify({
                "status": "error",
                "message": f"通信エラー: {str(e)}"
            }), 500
            
    except Exception as e:
        return jsonify({
            "status": "error", 
            "message": "サーバーエラー",
            "error": str(e)
        }), 500

@app.route('/angle_feedback', methods=['POST'])
def angle_feedback():
    try:
        data = request.get_json()
        servo1 = data.get("servo1", 25)
        servo2 = data.get("servo2", 17)
        servo3 = data.get("servo3", 137)
        servo4 = data.get("servo4", 120)
        arduino_response = data.get("arduino_response")
        
        return jsonify({
            "status": "success",
            "message": "フィードバックを受信しました"
        }), 200
    except Exception as e:
        return jsonify({
            "status": "error", 
            "message": str(e)
        }), 500


def verify_paths():
    """パスの存在を確認"""
    global ssh_client
    
    try:
        # Windowsのパス確認
        if not os.path.exists(VENV_PYTHON):
            raise FileNotFoundError(f"Python仮想環境が見つかりません: {VENV_PYTHON}")
        
        main_script_path = get_main_script_path()
        if not os.path.exists(main_script_path):
            raise FileNotFoundError(f"main.pyが見つかりません: {main_script_path}")
        
        # Raspberry Piのパス確認
        try:
            ssh_client = paramiko.SSHClient()
            ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            logger.info(f"Raspberry Piに接続を試みます: {RASPI_HOST}")
            ssh_client.connect(
                RASPI_HOST, 
                port=RASPI_PORT, 
                username=RASPI_USER, 
                password=RASPI_PASSWORD,
                timeout=10
            )
            
            # カメラスクリプトの存在確認
            stdin, stdout, stderr = ssh_client.exec_command(f"test -f {RASPI_CAMERA_SCRIPT} && echo 'exists'")
            result = stdout.read().decode().strip()
            if 'exists' not in result:
                raise FileNotFoundError(f"camera.pyが見つかりません: {RASPI_CAMERA_SCRIPT}")

            # シリアル通信スクリプトの存在確認
            stdin, stdout, stderr = ssh_client.exec_command(f"test -f {RASPI_SERIAL_SCRIPT} && echo 'exists'")
            result = stdout.read().decode().strip()
            if 'exists' not in result:
                raise FileNotFoundError(f"serial_angle.pyが見つかりません: {RASPI_SERIAL_SCRIPT}")
                
        except paramiko.SSHException as e:
            raise ConnectionError(f"SSH接続エラー: {e}")
        except socket.timeout:
            raise ConnectionError("接続がタイムアウトしました")
        except Exception as e:
            raise ConnectionError(f"Raspberry Piへの接続エラー: {e}")
            
        logger.info("全てのパスの確認が完了しました")
        return True
        
    except Exception as e:
        if ssh_client:
            ssh_client.close()
        raise e

def run_script(script_path, script_name):
    """指定したスクリプトを仮想環境のPythonで新しいプロセスとして実行"""
    logger.info(f"{script_name} を仮想環境で起動します...")
    try:
        # 完全なパスを構築
        full_script_path = os.path.abspath(os.path.join(script_path, script_name))
        working_dir = os.path.dirname(full_script_path)
        
        # パス情報をログ出力
        logger.info(f"スクリプトパス: {full_script_path}")
        logger.info(f"作業ディレクトリ: {working_dir}")
        
        # 環境変数を設定
        env = os.environ.copy()
        env['PYTHONUNBUFFERED'] = '1'  # Pythonの出力バッファリングを無効化
        env['PYTHONPATH'] = working_dir
        
        # プロセス生成時のフラグを設定
        creation_flags = 0
        if os.name == 'nt':
            creation_flags = subprocess.CREATE_NEW_PROCESS_GROUP

        # プロセスを起動
        process = subprocess.Popen(
            [VENV_PYTHON, full_script_path],
            creationflags=creation_flags,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=1,  # ラインバッファリングを設定
            universal_newlines=True,
            cwd=working_dir,
            env=env,
            shell=False  # シェル経由での実行を無効化
        )
        
        processes.append(process)
        logger.info(f"{script_name} を起動しました。PID: {process.pid}")
        
        # 出力を非同期で読み取るスレッドを作成
        def read_output(pipe, prefix):
            while True:
                line = pipe.readline()
                if not line:
                    break
                logger.info(f"{prefix}: {line.strip()}")

        stdout_thread = threading.Thread(target=read_output, args=(process.stdout, f"{script_name} [stdout]"))
        stderr_thread = threading.Thread(target=read_output, args=(process.stderr, f"{script_name} [stderr]"))
        
        stdout_thread.daemon = True
        stderr_thread.daemon = True
        stdout_thread.start()
        stderr_thread.start()
        
        return process
        
    except Exception as e:
        logger.error(f"{script_name} の起動に失敗しました: {e}")
        logger.exception(e)  # スタックトレースを出力
        return None
def monitor_process(process, name):
    """プロセスの出力を監視し、終了を検知する"""
    logger.info(f"プロセス {name} の監視を開始")
    
    while not shutdown_event.is_set():
        try:
            # プロセスの状態を確認
            if process.poll() is not None:
                exit_code = process.returncode
                logger.warning(f"{name} が終了しました。終了コード: {exit_code}")
                
                if exit_code != 0:
                    logger.error(f"{name} が異常終了しました")
                    # プロセスの出力バッファを最後まで読み取る
                    stdout, stderr = process.communicate()
                    if stdout:
                        logger.error(f"{name} の最終出力: {stdout.decode('utf-8', errors='ignore')}")
                    if stderr:
                        logger.error(f"{name} のエラー出力: {stderr.decode('utf-8', errors='ignore')}")
                break
            
            # プロセスが生きているか定期的に確認
            try:
                os.kill(process.pid, 0)
            except OSError:
                logger.error(f"{name} のプロセスが応答しません")
                break
                
            time.sleep(1)
            
        except Exception as e:
            logger.error(f"プロセス監視中にエラー: {e}")
            logger.exception(e)
            break

    logger.info(f"プロセス {name} の監視を終了")
def start_raspi_processes():
    """Raspberry Pi上でプロセスを起動"""
    global ssh_client, raspi_pid
    try:
        if ssh_client is None:
            ssh_client = paramiko.SSHClient()
            ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        
        if not ssh_client.get_transport() or not ssh_client.get_transport().is_active():
            ssh_client.connect(RASPI_HOST, port=RASPI_PORT, username=RASPI_USER, password=RASPI_PASSWORD)

        # 既存のプロセスを確実に終了
        logger.info("既存のプロセスを終了します...")
        ssh_client.exec_command("sudo pkill -f camera.py")
        ssh_client.exec_command("sudo pkill -f serial_angle.py")
        ssh_client.exec_command("sudo pkill -f mjpg_streamer")
        time.sleep(2)

        # プロセスが完全に終了したことを確認
        stdin, stdout, stderr = ssh_client.exec_command("pgrep -f 'camera.py|serial_angle.py|mjpg_streamer'")
        if stdout.read().decode().strip():
            logger.warning("既存のプロセスが残っています。強制終了を試みます。")
            ssh_client.exec_command("sudo pkill -9 -f 'camera.py|serial_angle.py|mjpg_streamer'")
            time.sleep(1)

        # カメラプロセスの起動
        logger.info("カメラプロセスを起動します...")
        ssh_client.exec_command(f"nohup python3 {RASPI_CAMERA_SCRIPT} > /tmp/camera.log 2>&1 &")
        time.sleep(2)

        # シリアル通信プロセスの起動
        logger.info("シリアル通信プロセスを起動します...")
        ssh_client.exec_command(f"nohup python3 {RASPI_SERIAL_SCRIPT} > /tmp/serial.log 2>&1 &")
        time.sleep(2)

        # プロセスIDの取得と確認
        camera_pid = get_remote_pid("camera.py")
        serial_pid = get_remote_pid("serial_angle.py")

        if camera_pid and serial_pid:
            logger.info(f"camera.py を起動しました。PID: {camera_pid}")
            logger.info(f"serial_angle.py を起動しました。PID: {serial_pid}")
            raspi_pid = {'camera': camera_pid, 'serial': serial_pid}

            # プロセスの動作確認
            check_serial_process()
            return True
        else:
            logger.error("一部のプロセスの起動に失敗しました")
            return False

    except Exception as e:
        logger.error(f"Raspberry Piプロセス起動時にエラー: {e}")
        if ssh_client:
            try:
                ssh_client.close()
            except:
                pass
        return False
def check_serial_process():
    """シリアル通信プロセスの動作確認"""
    try:
        response = requests.get(f"http://{RASPI_HOST}:5000/health", timeout=5)
        if response.status_code == 200:
            logger.info("シリアル通信プロセスが正常に動作しています")
            return True
        else:
            logger.warning("シリアル通信プロセスの応答が異常です")
            return False
    except requests.RequestException as e:
        logger.error(f"シリアル通信プロセスの確認中にエラー: {e}")
        return False

def get_remote_pid(script_name):
    """Raspberry Pi上の指定したスクリプトのPIDを取得"""
    global ssh_client
    try:
        stdin, stdout, stderr = ssh_client.exec_command(f"pgrep -f {script_name}")
        pid = stdout.read().decode().strip()
        if pid:
            logger.info(f"{script_name}のPID: {pid}")
        else:
            logger.warning(f"{script_name}のPIDが見つかりません")
        return pid
    except Exception as e:
        logger.error(f"PID取得中にエラー: {e}")
        return None

def emit_detection_status(status):
    """
    検知状態の変更をクライアントに通知
    """
    try:
        socketio.emit('detection_update', status)
    except Exception as e:
        logger.error(f"Error emitting detection status: {e}")

def cleanup_processes():
    """すべてのプロセスを終了"""
    global ssh_client, raspi_pid, cleanup_done, processes
    
    if cleanup_done:
        return
    
    logger.info("クリーンアップを開始します...")
    cleanup_done = True
    shutdown_event.set()

    try:
        # まずWindowsのプロセス（main.pyなど）を終了
        windows_processes = [p for p in processes if p and p.poll() is None]
        for process in windows_processes:
            try:
                pid = process.pid
                logger.info(f"プロセス (PID: {pid}) の終了を開始します...")
                
                # CTRL_BREAK_EVENTを送信して正常終了を試みる
                if os.name == 'nt':
                    try:
                        os.kill(pid, signal.CTRL_BREAK_EVENT)
                        # 終了を待機
                        process.wait(timeout=5)
                        logger.info(f"プロセス (PID: {pid}) が正常に終了しました")
                        continue
                    except (subprocess.TimeoutExpired, Exception) as e:
                        logger.warning(f"プロセス (PID: {pid}) の正常終了に失敗: {e}")

                # 強制終了を試みる
                try:
                    subprocess.run(['taskkill', '/F', '/T', '/PID', str(pid)], 
                                 capture_output=True,
                                 timeout=5)
                    process.wait(timeout=3)
                    logger.info(f"プロセス (PID: {pid}) を強制終了しました")
                except Exception as e:
                    logger.error(f"プロセス (PID: {pid}) の強制終了に失敗: {e}")
                
            except Exception as e:
                logger.error(f"プロセス終了中にエラー: {e}")

        # 次にRaspberry Pi上のプロセスを終了
        if ssh_client and raspi_pid:
            try:
                logger.info("Raspberry Pi上のプロセスを終了します...")
                
                # カメラプロセスの終了
                ssh_client.exec_command("sudo pkill -TERM -f camera.py")
                ssh_client.exec_command("sudo pkill -TERM -f mjpg_streamer")
                time.sleep(2)
                
                # シリアル通信プロセスの終了
                ssh_client.exec_command("sudo pkill -TERM -f serial_angle.py")
                time.sleep(2)
                
                # 残存プロセスの強制終了
                ssh_client.exec_command("sudo pkill -9 -f 'camera.py|serial_angle.py|mjpg_streamer'")
                
                logger.info("Raspberry Pi上のプロセスを終了しました")
            except Exception as e:
                logger.error(f"Raspberry Piプロセス終了時にエラー: {e}")
            finally:
                try:
                    ssh_client.close()
                    logger.info("SSH接続を閉じました")
                except:
                    pass

        # 最後にFlaskとSocketIOを停止（アプリケーションコンテキスト内で実行）
        try:
            def stop_server():
                logger.info("Flaskサーバーを停止します...")
                # SocketIOサーバーの停止
                socketio.stop()

            # アプリケーションコンテキスト内でFlask/SocketIOの停止を実行
            with app.app_context():
                stop_server()
                
        except Exception as e:
            logger.error(f"Flaskサーバー停止時にエラー: {e}")

    except Exception as e:
        logger.error(f"クリーンアップ中に予期せぬエラー: {e}")
    finally:
        logger.info("クリーンアップを完了しました")
        
        # プログラムを確実に終了
        try:
            # 少し待機して、終了処理が完了するのを待つ
            time.sleep(2)
            # システムレベルでの終了を実行
            os._exit(0)
        except:
            sys.exit(0)

def signal_handler(signum, frame):
    """シグナルハンドラー"""
    logger.info(f"シグナル {signum} を受信しました。終了処理を開始します...")
    
    try:
        # Flaskアプリケーションコンテキストの確保
        with app.app_context():
            # クリーンアップ処理を実行
            cleanup_processes()
    except Exception as e:
        logger.error(f"シグナルハンドラーでエラー: {e}")
        # エラーが発生した場合でも強制終了を試みる
        os._exit(1)

def main():
    """メイン実行関数"""
    try:
        # シグナルハンドラの設定
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

        # パスの検証
        logger.info("パスを確認中...")
        verify_paths()
        logger.info("パスの確認が完了しました")

        # Raspberry Piのプロセスを起動
        logger.info("Raspberry Piのプロセスを起動します...")
        if not start_raspi_processes():
            logger.error("Raspberry Piのプロセス起動に失敗しました")
            cleanup_processes()
            return

        # main.pyを起動
        logger.info("main.pyを起動します...")
        main_process = run_script(WINDOWS_MAIN_PATH, WINDOWS_MAIN_SCRIPT)
        if not main_process:
            logger.error("main.pyの起動に失敗しました")
            cleanup_processes()
            return

        # プロセス監視関数
        def monitor_main_process():
            try:
                while not shutdown_event.is_set():
                    if main_process.poll() is not None:
                        exit_code = main_process.poll()
                        logger.error(f"main.py が終了しました。終了コード: {exit_code}")
                        if not cleanup_done:
                            cleanup_processes()
                        break
                    time.sleep(1)
            except Exception as e:
                logger.error(f"プロセス監視中にエラー: {e}")
                if not cleanup_done:
                    cleanup_processes()

        # 監視スレッドの開始
        monitor_thread = Thread(target=monitor_main_process)
        monitor_thread.daemon = True
        monitor_thread.start()

        # Flaskサーバーの起動
        logger.info(f"Flaskサーバーを起動します (port: {FLASK_PORT})...")
        try:
            socketio.run(
                app,
                host="0.0.0.0",
                port=FLASK_PORT,
                allow_unsafe_werkzeug=True,
                use_reloader=False
            )
        except KeyboardInterrupt:
            logger.info("キーボード割り込みを検知しました")
            cleanup_processes()
        except Exception as e:
            logger.error(f"Flaskサーバー実行中にエラー: {e}")
            cleanup_processes()

    except Exception as e:
        logger.error(f"予期せぬエラーが発生しました: {e}")
        cleanup_processes()
    finally:
        if not cleanup_done:
            cleanup_processes()

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        logger.info("KeyboardInterruptを受信しました")
        if not cleanup_done:
            cleanup_processes()
    except Exception as e:
        logger.error(f"予期せぬエラーが発生しました: {e}")
        if not cleanup_done:
            cleanup_processes()
        sys.exit(1)