// API設定
const API_CONFIG = {
    MAIN_SERVER: 'http://172.25.13.131:9000',     // WindowsPC
    CONTROL_SERVER: 'http://172.25.13.131:9000'   // WindowsPC
};

// グローバル変数の定義
window.isRecording = false;
window.recordingTimer = null;
window.recordingStartTime = null;

// WebSocket設定
let socket;
let retryCount = 0;
const MAX_RETRY = 5;

// カメラ制御関連の変数
let selectedCamera = 1; // デフォルトは前方カメラ
let horizontalAngle = 0;
let verticalAngle = 0;
let keyPressed = {};
let sendQueue = [];
const SEND_INTERVAL = 100;

// 初期化関数
function initializeApplication() {
    connectWebSocket();
    startDetectionUpdates();
    setInterval(sendAngleData, SEND_INTERVAL);

    // イベントリスナーの設定
    document.addEventListener('keydown', handleKeyDown);
    document.addEventListener('keyup', handleKeyUp);
}

// エラー処理関数
function showError(message) {
    const errorElement = document.getElementById('errorMessage');
    if (errorElement) {
        errorElement.textContent = message;
    }
    console.error(message);
}

function clearError() {
    const errorElement = document.getElementById('errorMessage');
    if (errorElement) {
        errorElement.textContent = '';
    }
}

// タブ切り替え機能
window.showTab = function(tabId) {
    const tabs = document.querySelectorAll('.tab');
    const headers = document.querySelectorAll('.tab-header');

    tabs.forEach(tab => {
        tab.classList.remove('active');
    });
    headers.forEach(header => {
        header.classList.remove('active');
    });

    document.getElementById(tabId).classList.add('active');
    document.querySelector(`.tab-header[data-tab="${tabId}"]`).classList.add('active');
};

// WebSocket接続
function connectWebSocket() {
    if (socket && socket.connected) {
        console.log('WebSocket already connected');
        return;
    }

    console.log('Attempting to connect to WebSocket...');
    socket = io(API_CONFIG.MAIN_SERVER, {
        transports: ['websocket'],
        reconnectionAttempts: MAX_RETRY,
        reconnectionDelay: 1000
    });

    socket.on('connect', () => {
        console.log('WebSocket connected successfully');
        retryCount = 0;
        clearError();
        startDetectionUpdates();
    });

    socket.on('disconnect', () => {
        console.warn("WebSocket接続が切断されました");
        if (retryCount < MAX_RETRY) {
            retryCount++;
            console.log(`再接続を試みます (${retryCount}/${MAX_RETRY})...`);
            setTimeout(connectWebSocket, 2000);
        } else {
            console.error("最大再試行回数に達しました");
            showError("サーバーへの接続が失われました。ページを再読み込みしてください。");
        }
    });

    socket.on('camera_stream_home_1', (data) => {
        updateCameraFrame('cameraStream1', data.image);
    });

    socket.on('camera_stream_home_2', (data) => {
        updateCameraFrame('cameraStream2', data.image);
    });

    socket.on('detection_update', (data) => {
        console.log('Received detection update:', data);
        updateDetectionStatus(data);
    });
}

// カメラストリーム関連の機能
function checkCameraStreams() {
    const stream1 = document.getElementById('cameraStream1');
    const stream2 = document.getElementById('cameraStream2');
    
    if (stream1) {
        stream1.onload = () => {
            console.log("カメラ1のストリームを受信中");
            clearError();
        };
    }
    
    if (stream2) {
        stream2.onload = () => {
            console.log("カメラ2のストリームを受信中");
            clearError();
        };
    }
}

function updateCameraFrame(elementId, imageData) {
    const imgElement = document.getElementById(elementId);
    if (imgElement) {
        imgElement.src = `data:image/jpeg;base64,${imageData}`;
    }
}

function handleStreamError(cameraNum) {
    console.error(`カメラ${cameraNum}の接続に失敗しました`);
    const streamElement = document.getElementById(`cameraStream${cameraNum}`);
    if (streamElement) {
        streamElement.src = "./static/css/no-video.png";
        showError(`カメラ${cameraNum}に接続できません。接続を確認してください。`);
    }
}

// 録画制御
// 録画制御のデバッグを強化
window.startRecording = function() {
    console.log('録画開始リクエスト送信');
    
    fetch(`${API_CONFIG.MAIN_SERVER}/start_recording`, {
        method: 'POST',
        headers: { 'Content-Type': 'application/json' }
    })
    .then(response => {
        console.log('録画開始レスポンス:', response.status); // ステータスコードを確認
        return response.json();
    })
    .then(data => {
        console.log('録画開始レスポンスデータ:', data); // レスポンスの内容を確認
        if (data.status === 'success') {
            window.isRecording = true;
            updateRecordingUI();
            window.recordingStartTime = Date.now();
            startRecordingTimer();
            console.log('録画を開始しました');
        } else {
            console.error('録画開始に失敗しました:', data.message);
            showError('録画開始に失敗しました');
        }
    })
    .catch(error => {
        console.error('録画開始中にエラー:', error);
        showError('録画開始中にエラーが発生しました');
    });
};
window.stopRecording = function() {
    console.log('録画停止リクエスト送信');
    
    fetch(`${API_CONFIG.MAIN_SERVER}/stop_recording`, {
        method: 'POST',
        headers: { 'Content-Type': 'application/json' }
    })
    .then(response => response.json())
    .then(data => {
        if (data.status === 'success') {
            window.isRecording = false;
            updateRecordingUI();
            stopRecordingTimer();
            console.log('録画を停止しました');
        } else {
            console.error('録画停止に失敗しました:', data.message);
            showError('録画停止に失敗しました');
        }
    })
    .catch(error => {
        console.error('録画停止中にエラー:', error);
        showError('録画停止中にエラーが発生しました');
    });
};

function updateRecordingUI() {
    const startButton = document.getElementById('startRecordButton');
    const stopButton = document.getElementById('stopRecordButton');
    const timeDisplay = document.getElementById('recordingTime');
    
    if (window.isRecording) {
        startButton.disabled = true;
        stopButton.disabled = false;
        timeDisplay.style.display = 'block';
    } else {
        startButton.disabled = false;
        stopButton.disabled = true;
        timeDisplay.style.display = 'none';
    }
}

function startRecordingTimer() {
    if (window.recordingTimer) {
        clearInterval(window.recordingTimer);
    }
    updateRecordingTime();
    window.recordingTimer = setInterval(updateRecordingTime, 1000);
}

function stopRecordingTimer() {
    if (window.recordingTimer) {
        clearInterval(window.recordingTimer);
        window.recordingTimer = null;
    }
    window.recordingStartTime = null;
}

function updateRecordingTime() {
    if (!window.recordingStartTime) return;
    
    const elapsed = Date.now() - window.recordingStartTime;
    const seconds = Math.floor(elapsed / 1000) % 60;
    const minutes = Math.floor(elapsed / (1000 * 60)) % 60;
    const hours = Math.floor(elapsed / (1000 * 60 * 60));
    
    const timeString = `${hours.toString().padStart(2, '0')}:${minutes.toString().padStart(2, '0')}:${seconds.toString().padStart(2, '0')}`;
    
    const timeElement = document.getElementById('recordingTime');
    if (timeElement) {
        timeElement.textContent = `録画時間: ${timeString}`;
    }
}

// 人物検知状態の管理
let detectionUpdateInterval;

function startDetectionUpdates() {
    checkDetectionStatus();
    detectionUpdateInterval = setInterval(checkDetectionStatus, 1000);
}

function stopDetectionUpdates() {
    if (detectionUpdateInterval) {
        clearInterval(detectionUpdateInterval);
    }
}

function checkDetectionStatus() {
    fetch(`${API_CONFIG.MAIN_SERVER}/detection_status`)
        .then(response => response.json())
        .then(data => {
            console.log('Detection status response:', data);
            updateDetectionStatus(data);
        })
        .catch(error => {
            console.error('検知状態の取得に失敗しました:', error);
        });
}

function updateDetectionStatus(data) {
    const updateElement = (elementId, detectionId, detected) => {
        const statusElement = document.getElementById(elementId);
        const detectionElement = document.getElementById(detectionId);
        
        if (statusElement) {
            const cameraName = elementId === 'camera1Status' ? '前方カメラ' : '後方カメラ';
            statusElement.textContent = `${cameraName}: ${detected ? '人物検出中' : '未検出'}`;
            statusElement.className = `status-indicator ${detected ? 'detected' : 'not-detected'}`;
        }
        
        if (detectionElement) {
            detectionElement.textContent = `人物: ${detected ? '検出' : '未検出'}`;
            detectionElement.className = `detection-overlay ${detected ? 'detected' : 'not-detected'}`;
        }
    };

    updateElement('camera1Status', 'detection1', data.camera1_detected);
    updateElement('camera2Status', 'detection2', data.camera2_detected);
}

// カメラ選択制御
window.selectCameraButton = function(buttonId) {
    const cameraButtons = document.querySelectorAll('.camera-button');
    cameraButtons.forEach(button => {
        button.classList.remove('selected');
        button.classList.add('not-selected');
    });

    const selectedButton = document.getElementById(buttonId);
    selectedButton.classList.add('selected');
    selectedButton.classList.remove('not-selected');

    selectedCamera = buttonId === 'cameraButton1' ? 1 : 2;
    console.log(`カメラ${selectedCamera}が制御対象として選択されました`);
};

// 角度制御
function updateAngleDisplay1() {
    document.getElementById('angle-display1').textContent = `左右角度: ${horizontalAngle}°`;
}

function updateAngleDisplay2() {
    document.getElementById('angle-display2').textContent = `上下角度: ${verticalAngle}°`;
}

function sendAngleData() {
    if (sendQueue.length === 0) return;
    
    const data = sendQueue.shift();
    data.selectedCamera = selectedCamera;

    fetch(`${API_CONFIG.CONTROL_SERVER}/update_angle`, {
        method: 'POST',
        headers: { 
            'Content-Type': 'application/json',
            'Access-Control-Allow-Origin': '*'
        },
        body: JSON.stringify(data)
    })
    .then(response => response.json())
    .then(data => {
        console.log(`サーバーからのレスポンス:`, data);
    })
    .catch(error => {
        console.error("送信中にエラーが発生しました: ", error);
        sendQueue.push(data);
    });
}

function enqueueAngleData(horizontal, vertical) {
    sendQueue.push({
        horizontal: horizontal,
        vertical: vertical
    });
}

// キーボード制御
function handleKeyDown(event) {
    if (!keyPressed[event.key]) {
        keyPressed[event.key] = true;
        handlePress(event.key);
    }
}

function handleKeyUp(event) {
    keyPressed[event.key] = false;
    const buttonMap = {
        "ArrowUp": "up-button",
        "ArrowDown": "down-button",
        "ArrowLeft": "left-button",
        "ArrowRight": "right-button"
    };
    
    if (buttonMap[event.key]) {
        document.getElementById(buttonMap[event.key]).classList.remove("dark");
    }
}

function handlePress(key) {
    let angleChanged = false;
    
    switch(key) {
        case "ArrowUp":
            if (verticalAngle < 90) {
                verticalAngle = Math.min(90, verticalAngle + 5);  // 5度ずつ変更
                angleChanged = true;
                document.getElementById("up-button").classList.add("dark");
            }
            break;
        case "ArrowDown":
            if (verticalAngle > -90) {
                verticalAngle = Math.max(-90, verticalAngle - 5);  // 5度ずつ変更
                angleChanged = true;
                document.getElementById("down-button").classList.add("dark");
            }
            break;
        case "ArrowLeft":
            if (horizontalAngle > -90) {
                horizontalAngle = Math.max(-90, horizontalAngle - 5);  // 5度ずつ変更
                angleChanged = true;
                document.getElementById("left-button").classList.add("dark");
            }
            break;
        case "ArrowRight":
            if (horizontalAngle < 90) {
                horizontalAngle = Math.min(90, horizontalAngle + 5);  // 5度ずつ変更
                angleChanged = true;
                document.getElementById("right-button").classList.add("dark");
            }
            break;
    }

    if (angleChanged) {
        enqueueAngleData(horizontalAngle, verticalAngle);
    }
}

function enqueueAngleData(horizontal, vertical) {
    // 左右の値をサーボ1と3の値に変換
    let servo1 = 25 - horizontal;  // 初期値25°からの変化
    let servo3 = 137 - horizontal; // 初期値137°からの変化（反転）
    
    // 上下の値をサーボ2と4の値に変換
    let servo2 = 17 + vertical;    // 初期値17°からの変化
    let servo4 = 120 + vertical;   // 初期値120°からの変化（反転）

    sendQueue.push({
        servo1: servo1,
        servo2: servo2,
        servo3: servo3,
        servo4: servo4
    });
}

// マウス/タッチ制御
window.handleMouseDown = function(event, direction) {
    event.preventDefault();
    let angleChanged = false;

    switch(direction) {
        case 'up':
            if (verticalAngle < 90) {
                verticalAngle += 1;
                angleChanged = true;
                updateAngleDisplay2();
            }
            break;
        case 'down':
            if (verticalAngle > -90) {
                verticalAngle -= 1;
                angleChanged = true;
                updateAngleDisplay2();
            }
            break;
        case 'left':
            if (horizontalAngle > -90) {
                horizontalAngle -= 1;
                angleChanged = true;
                updateAngleDisplay1();
            }
            break;
        case 'right':
            if (horizontalAngle < 90) {
                horizontalAngle += 1;
                angleChanged = true;
                updateAngleDisplay1();
            }
            break;
    }

    if (angleChanged) {
        enqueueAngleData(horizontalAngle, verticalAngle);
    }
    
    document.getElementById(event.target.id).classList.add('dark');
    window.intervalId = setInterval(() => {
        handleMousePress(direction);
    }, 50);
};

window.handleMousePress = function(direction) {
    let angleChanged = false;

    switch(direction) {
        case 'up':
            if (verticalAngle < 90) {
                verticalAngle += 1;
                angleChanged = true;
                updateAngleDisplay2();
            }
            break;
        case 'down':
            if (verticalAngle > -90) {
                verticalAngle -= 1;
                angleChanged = true;
                updateAngleDisplay2();
            }
            break;
        case 'left':
            if (horizontalAngle > -90) {
                horizontalAngle -= 1;
                angleChanged = true;
                updateAngleDisplay1();
            }
            break;
        case 'right':
            if (horizontalAngle < 90) {
                horizontalAngle += 1;
                angleChanged = true;
                updateAngleDisplay1();
            }
            break;
    }

    if (angleChanged) {
        enqueueAngleData(horizontalAngle, verticalAngle);
    }
};

window.handleMouseUp = function(event) {
    if (window.intervalId) {
        clearInterval(window.intervalId);
        window.intervalId = null;
    }
    document.getElementById(event.target.id).classList.remove('dark');
};

// 初期化時の処理
document.addEventListener('DOMContentLoaded', () => {
    // タブの初期表示
    showTab('home');

    // アニメーション遷移の設定
    const animationScreen = document.getElementById('animation');
    const mainScreen = document.getElementById('main');
    
    if (animationScreen && mainScreen) {
        animationScreen.addEventListener('click', () => {
            animationScreen.style.display = 'none';
            mainScreen.style.display = 'block';
        });
    }

    // ボタンのイベントリスナーを設定
    initializeRecordingButtons();

    // 各機能の初期化
    initializeApplication();
    initializeVolumeControl();
});

// 録画ボタンの初期化
function initializeRecordingButtons() {
    const startButton = document.getElementById('startRecordButton');
    const stopButton = document.getElementById('stopRecordButton');

    if (startButton && stopButton) {
        updateRecordingUI();  // 初期状態の設定
    }
}

// 音量制御の初期化
function initializeVolumeControl() {
    const volumeSlider = document.getElementById('volumeSlider');
    const video = document.querySelector('.video2');
    
    if (volumeSlider && video) {
        volumeSlider.addEventListener('input', () => {
            video.volume = volumeSlider.value;
        });
        video.volume = volumeSlider.value;
    }
}

// クリーンアップ処理
window.addEventListener('beforeunload', (event) => {
    if (window.isRecording) {
        stopRecording();
    }
    stopDetectionUpdates();
    if (socket) {
        socket.disconnect();
    }
});

// エクスポート（モジュール環境の場合）
if (typeof module !== 'undefined' && module.exports) {
    module.exports = {
        startRecording,
        stopRecording,
        selectCameraButton,
        handleMouseDown,
        handleMouseUp,
        showTab
    };
}