import cv2
import numpy as np
import urllib.request
import threading
import time
import os
import logging
import requests 
import signal
from datetime import datetime

def setup_logging():
    """
    ロギング設定を初期化する詳細な関数
    """
    # ログディレクトリの作成
    log_dir = os.path.join(os.path.dirname(__file__), 'logs')
    
    try:
        # ディレクトリが存在しない場合は作成
        os.makedirs(log_dir, exist_ok=True)
        
        # ログファイルのパス設定（絶対パスを使用）
        log_file = os.path.join(log_dir, 'main.log')
        
        # ロガーを取得（__name__を使用）
        logger = logging.getLogger(__name__)
        
        # ロガーをクリア（重複ハンドラを防ぐ）
        for handler in logger.handlers[:]:
            logger.removeHandler(handler)
        
        # ロガーのレベル設定
        logger.setLevel(logging.INFO)
        
        # フォーマッターの作成
        formatter = logging.Formatter(
            '%(asctime)s - %(levelname)s - [%(filename)s:%(lineno)d] - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        )
        
        # ファイルハンドラ（UTF-8エンコーディング、BOMなし）
        try:
            file_handler = logging.FileHandler(log_file, encoding='utf-8')
            file_handler.setLevel(logging.INFO)
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)
        except PermissionError:
            print(f"警告: ログファイル {log_file} への書き込み権限がありません。")
        except Exception as e:
            print(f"ファイルハンドラの作成中にエラーが発生しました: {e}")
        
        # コンソールハンドラ
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
        
        # ロギングの開始を記録
        logger.info("=" * 50)
        logger.info("ロギングシステムを初期化しました")
        logger.info(f"ログファイルのパス: {log_file}")
        logger.info("=" * 50)
    
    except Exception as e:
        print(f"ロギングシステムの初期化中に重大なエラーが発生しました: {e}")
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[logging.StreamHandler()]
        )
    return logger

logger = logging.getLogger(__name__)

class PersonDetector:
    def __init__(self):
        """人物検知機能の初期化"""
        self.hog = cv2.HOGDescriptor()
        self.hog.setSVMDetector(cv2.HOGDescriptor_getDefaultPeopleDetector())
        self.detection_lock = threading.Lock()
        self.logger = logging.getLogger(__name__)

    def detect_persons(self, frame):
        """
        フレーム内の人物を検出（スレッドセーフ）
        """
        if frame is None:
            return False
            
        try:
            with self.detection_lock:
                # フレームのコピーを作成して処理（元のフレームを変更しない）
                frame_copy = frame.copy()
                
                # フレームのサイズを調整（処理速度向上のため）
                height, width = frame_copy.shape[:2]
                scale = min(640.0 / width, 480.0 / height)
                if scale < 1:
                    frame_copy = cv2.resize(frame_copy, None, fx=scale, fy=scale)

                # 人物検出の実行
                boxes, weights = self.hog.detectMultiScale(
                    frame_copy,
                    winStride=(8, 8),
                    padding=(8, 8),
                    scale=1.05,
                    hitThreshold=0
                )

                return len(boxes) > 0 and any(weight > 0.3 for weight in weights)

        except Exception as e:
            self.logger.error(f"人物検知処理中にエラー: {e}")
            return False
        
class RecordingManager:
    def __init__(self, record_dir: str):
        self.record_dir = record_dir
        self._recording_buffer = [[], []]  # 実際の録画用バッファ
        self._segment_start_time = None
        self._last_frame_time = [None, None]
        self._min_segment_frames = 550  # 最小フレーム数（55秒分@10fps）
        self.target_fps = 10
        self.lock = threading.Lock()
        self.logger = logging.getLogger(__name__)

    def add_frame(self, frame: np.ndarray, camera_id: int) -> None:
        """フレームをバッファに追加し、必要に応じてセグメントを保存"""
        with self.lock:
            current_time = time.time()
            
            # セグメント開始時間の初期化
            if self._segment_start_time is None:
                self._segment_start_time = current_time
                self.logger.info(f"Starting new segment for camera {camera_id}")

            # フレームの追加
            self._recording_buffer[camera_id].append(frame.copy())
            self._last_frame_time[camera_id] = current_time

            # バッファサイズとフレームレートの管理
            self._manage_buffer(camera_id)

            # セグメント終了チェック
            if self._should_save_segment(camera_id, current_time):
                self._save_segment(camera_id)

    def _manage_buffer(self, camera_id: int) -> None:
        """バッファサイズとフレームレートを管理"""
        if len(self._recording_buffer[camera_id]) > self.target_fps * 65:  # 余裕を持って65秒分
            # フレームの間引き処理
            target_frame_count = self.target_fps * 60  # 60秒分を維持
            indices = np.linspace(0, len(self._recording_buffer[camera_id]) - 1, target_frame_count, dtype=int)
            self._recording_buffer[camera_id] = [self._recording_buffer[camera_id][i] for i in indices]

    def _should_save_segment(self, camera_id: int, current_time: float) -> bool:
        """セグメントを保存すべきかどうかを判断"""
        if self._segment_start_time is None:
            return False

        elapsed_time = current_time - self._segment_start_time
        has_min_frames = len(self._recording_buffer[camera_id]) >= self._min_segment_frames
        return elapsed_time >= 60.0 and has_min_frames

    def _save_segment(self, camera_id: int) -> None:
        """セグメントを動画ファイルとして保存"""
        if not self._recording_buffer[camera_id]:
            self.logger.warning(f"No frames to save for camera {camera_id}")
            return

        try:
            # ファイル名の生成
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            filename = os.path.join(
                self.record_dir,
                f"camera{camera_id+1}_{timestamp}_60sec.mp4"
            )

            # フレーム情報の取得
            frames = self._recording_buffer[camera_id]
            frame_count = len(frames)
            actual_duration = (self._last_frame_time[camera_id] - self._segment_start_time)
            actual_fps = frame_count / actual_duration if actual_duration > 0 else self.target_fps

            # VideoWriterの設定
            frame_size = frames[0].shape[1::-1]
            fourcc = cv2.VideoWriter_fourcc(*'mp4v')
            writer = cv2.VideoWriter(filename, fourcc, actual_fps, frame_size)

            # フレームの書き込み
            for frame in frames:
                writer.write(frame)

            writer.release()

            self.logger.info(
                f"Camera {camera_id}: Saved segment with {frame_count} frames "
                f"at {actual_fps:.2f} fps (duration: {actual_duration:.2f}s)"
            )

        except Exception as e:
            self.logger.error(f"Error saving segment for camera {camera_id}: {str(e)}")

        finally:
            # バッファのリセット
            self._recording_buffer[camera_id] = []
            self._segment_start_time = None
            self._last_frame_time[camera_id] = None
            
class DualCameraProcessor:
    def __init__(self, stream_urls=None, record_dir=None):
        """
        DualCameraProcessorの初期化
        Args:
            stream_urls (list): カメラストリームのURL一覧。Noneの場合はデフォルト値を使用
            record_dir (str): 録画ファイルの保存ディレクトリ。Noneの場合はデフォルトパスを使用
        """
        # ストリームURL
        self.stream_urls = stream_urls or [
            "http://172.25.16.204:8080/?action=stream",  # カメラ1
            "http://172.25.16.204:8081/?action=stream"   # カメラ2
        ]
        
        # 録画ディレクトリの設定
        if record_dir is None:
            base_dir = os.path.dirname(os.path.abspath(__file__))
            self.record_dir = os.path.join(base_dir, "recordings")
        else:
            self.record_dir = record_dir
        
        # 録画ディレクトリの作成と権限確認
        try:
            os.makedirs(self.record_dir, exist_ok=True)
            # ディレクトリの書き込み権限を確認
            test_file = os.path.join(self.record_dir, "test.txt")
            with open(test_file, "w") as f:
                f.write("test")
            os.remove(test_file)
            logging.info(f"録画ディレクトリを確認: {self.record_dir}")
        except Exception as e:
            logging.error(f"録画ディレクトリの設定中にエラー: {e}")
            raise
        
        # フレーム処理設定
        self.frames = [None, None]
        self.lock = threading.Lock()
        
        # 検知状態
        self.detection_status = {
            'camera1_detected': False,
            'camera2_detected': False
        }
        
        # 録画設定
        self.recording_manager = RecordingManager(self.record_dir)
        self.auto_recording_active = False
        
        # シャットダウン制御用のイベントを追加
        self.shutdown_event = threading.Event()
        
        # スレッド管理
        self.threads = []
        self.running = True

        # 人物検知関連の設定
        self.person_detector = PersonDetector()
        self.detection_interval = 1.0  # 検出間隔（秒）
        self.last_detection_time = [0, 0]  # 各カメラの最後の検出時刻
        
        logging.info("DualCameraProcessorの初期化が完了しました")
        logging.info(f"録画ディレクトリ: {self.record_dir}")
        logging.info(f"ストリームURL: {self.stream_urls}")

    def start_streaming(self):
        """すべてのカメラのストリーミングを開始"""
        try:
            # ストリーミングスレッドの開始
            for i, url in enumerate(self.stream_urls):
                thread = threading.Thread(target=self._process_stream, args=(url, i))
                thread.daemon = True
                thread.start()
                self.threads.append(thread)
                logger.info(f"カメラ {i+1} のストリーミングを開始")
            return True
        except Exception as e:
            logger.error(f"ストリーミング開始中にエラー: {e}")
            return False
        
    def _process_stream(self, url, camera_id):
        """各カメラのストリーム処理"""
        stream = None
        bytes_data = bytes()
        
        while self.running and not self.shutdown_event.is_set():
            try:
                if stream is None:
                    stream = urllib.request.urlopen(url)
                
                chunk = stream.read(1024)
                if not chunk:
                    raise ConnectionError("ストリームが切断されました")
                
                bytes_data += chunk
                a = bytes_data.find(b'\xff\xd8')
                b = bytes_data.find(b'\xff\xd9')
                
                if a != -1 and b != -1:
                    jpg = bytes_data[a:b+2]
                    bytes_data = bytes_data[b+2:]
                    
                    frame = cv2.imdecode(np.frombuffer(jpg, dtype=np.uint8), cv2.IMREAD_COLOR)
                    if frame is not None:
                        with self.lock:
                            # 現在のフレームを更新
                            self.frames[camera_id] = frame
                            
                            # 録画処理
                            if self.auto_recording_active and camera_id ==0:
                                self.recording_manager.add_frame(frame, camera_id)

                            # 人物検知処理
                            current_time = time.time()
                            if current_time - self.last_detection_time[camera_id] >= self.detection_interval:
                                is_detected = self.person_detector.detect_persons(frame)
                                
                                # 検知状態が変化した場合のみ通知
                                previous_state = self.detection_status[f'camera{camera_id+1}_detected']
                                if is_detected != previous_state:
                                    if is_detected:
                                        logger.info(f"カメラ {camera_id+1} で人物を検出しました")
                                    else:
                                        logger.info(f"カメラ {camera_id+1} の人物検知が終了しました")
                                    
                                    # 検知状態の更新
                                    self.detection_status[f'camera{camera_id+1}_detected'] = is_detected
                                    
                                    # 状態更新をサーバーに送信
                                    try:
                                        response = requests.post(
                                            'http://localhost:9000/detection_status',
                                            json=self.detection_status,
                                            headers={'Content-Type': 'application/json'},
                                            timeout=1
                                        )
                                        if response.status_code != 200:
                                            logger.error(f"検知状態の送信失敗: {response.status_code}")
                                    except Exception as e:
                                        logger.error(f"検知状態の送信に失敗: {e}")
                                
                                self.last_detection_time[camera_id] = current_time

            except Exception as e:
                logger.error(f"カメラ {camera_id} の処理中にエラー: {e}")
                if stream:
                    try:
                        stream.close()
                    except:
                        pass
                    stream = None
                if not self.shutdown_event.is_set():
                    time.sleep(1)
                    continue
                break

        if stream:
            stream.close()

    def start_auto_recording(self):
        """自動録画の開始"""
        if self.auto_recording_active:
            logger.warning("自動録画は既に実行中です")
            return False

        self.auto_recording_active = True
        logger.info("自動録画を開始")
        return True

    def stop_auto_recording(self):
        """自動録画の停止"""
        logger.info("自動録画の停止を開始...")
        self.auto_recording_active = False
        logger.info("自動録画を停止しました")

    def shutdown(self):
        """プロセッサーの完全シャットダウン"""
        logger.info("DualCameraProcessorのシャットダウンを開始...")
        
        # シャットダウンイベントをセット
        self.shutdown_event.set()
        
        # 録画停止
        self.stop_auto_recording()
        
        # ストリーミング停止
        self.running = False
        
        # すべてのスレッドの終了を待機
        for thread in self.threads:
            if thread.is_alive():
                thread.join(timeout=5)
        
        logger.info("DualCameraProcessorのシャットダウンが完了しました")

def main():
    # ロギング設定
    setup_logging()
    logger = logging.getLogger(__name__)
    
    processor = None
    
    def signal_handler(signum, frame):
        """シグナルハンドラー"""
        logger.info(f"シグナル {signum} を受信しました")
        nonlocal processor
        if processor:
            processor.shutdown()
        
    # シグナルハンドラーの設定
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    try:
        processor = DualCameraProcessor()
        
        if not processor.start_streaming():
            logger.error("ストリーミングの開始に失敗しました")
            return
        
        if not processor.start_auto_recording():
            logger.error("自動録画の開始に失敗しました")
            processor.shutdown()
            return
        
        # メインループ
        try:
            while not processor.shutdown_event.is_set():
                time.sleep(1)
        except KeyboardInterrupt:
            logger.info("ユーザーによる中断を検知")
            
    except Exception as e:
        logger.error(f"アプリケーション実行中にエラーが発生: {e}")
    finally:
        if processor:
            processor.shutdown()

if __name__ == "__main__":
    main()