from flask import Flask, render_template, jsonify, request
import socket
import json
import threading
import time
from datetime import datetime
import sys

# 目的地設定
from destinations import DESTINATIONS, ROUTES

app = Flask(__name__)

class VehicleTracker:
    def __init__(self, host='0.0.0.0', port=5001):
        self.host = host
        self.port = port
        self.current_position = {"x": 50.5, "y": 5.5}  # 初期位置
        
        self.destinations = DESTINATIONS
        self.routes = ROUTES
        
        self.current_route = None
        
        # クライアント管理
        self.clients = {}
        self.clients_lock = threading.Lock()
        
        # 位置情報の保護 
        self.position_lock = threading.Lock()
        
        # サーバーソケット
        self.server_socket = None
        self.is_running = True
        
        # デバッグ用のフラグ
        self.debug = True

    def update_position(self, new_position):
        """位置情報の更新と記録"""
        with self.position_lock:
            old_position = dict(self.current_position)
            self.current_position = new_position
            if self.debug:
                print(f"位置情報を更新:")
                print(f"  前: {old_position}")
                print(f"  後: {self.current_position}")
                print(f"  更新時刻: {datetime.now().strftime('%H:%M:%S')}")
        return True

    def handle_client(self, conn, addr):
        """クライアント接続の処理"""
        client_id = f"{addr[0]}:{addr[1]}"
        
        with self.clients_lock:
            self.clients[client_id] = {
                "socket": conn,
                "last_seen": time.time(),
                "address": addr
            }
        
        print(f"クライアント接続: {addr}")
        
        # 初期目的地を送信
        if self.current_route:
            try:
                initial_destination = json.dumps({
                    "type": "destination",
                    "data": self.routes[self.current_route][0],
                    "timestamp": time.time()
                })
                conn.send(initial_destination.encode())
                print(f"初期目的地を送信: {self.routes[self.current_route][0]}")
            except Exception as e:
                print(f"初期目的地送信エラー: {e}")
        
        while self.is_running:
            try:
                conn.settimeout(30)
                data = conn.recv(1024).decode()
                if not data:
                    break
                
                # 複数のJSONメッセージを処理
                current_time = time.time()
                messages = self.parse_messages(data)
                
                for message in messages:
                    try:
                        with self.clients_lock:
                            if client_id in self.clients:
                                self.clients[client_id]["last_seen"] = current_time
                        
                        if message["type"] == "position":
                            position_data = message["data"]
                            if self.update_position(position_data):
                                # 確認応答を送信
                                response = json.dumps({
                                    "type": "position_ack",
                                    "timestamp": time.time()
                                })
                                conn.send(response.encode())
                                print(f"位置情報を受信: {position_data}")
                                
                    except Exception as e:
                        print(f"メッセージ処理エラー: {e}")
                        continue
                
            except socket.timeout:
                print(f"クライアント {client_id} の応答タイムアウト")
                break
            except Exception as e:
                print(f"通信エラー: {e}")
                break
        
        self.cleanup_client(client_id)

    def parse_messages(self, data):
        """複数のJSONメッセージを解析"""
        messages = []
        try:
            # データがJSON配列の場合の処理
            if data.startswith('[') and data.endswith(']'):
                return json.loads(data)
            
            # 連結されたJSONメッセージの処理
            for message in data.strip().split('}{'):
                try:
                    # メッセージの修復
                    if not message.startswith('{'):
                        message = '{' + message
                    if not message.endswith('}'):
                        message = message + '}'
                    
                    parsed = json.loads(message)
                    messages.append(parsed)
                    if self.debug:
                        print(f"メッセージを解析: {parsed}")
                except json.JSONDecodeError as e:
                    print(f"メッセージ解析エラー: {message}")
                    print(f"エラー詳細: {str(e)}")
                    continue
        except Exception as e:
            print(f"メッセージ解析中の予期せぬエラー: {str(e)}")
        
        return messages

    def cleanup_client(self, client_id):
        """クライアント接続のクリーンアップ"""
        with self.clients_lock:
            if client_id in self.clients:
                try:
                    self.clients[client_id]["socket"].close()
                except:
                    pass
                del self.clients[client_id]
                print(f"クライアント {client_id} の接続を終了しました")

    def broadcast_single_destination(self, destination_name):
        """単一の目的地を送信（経由点を含む場合はリストとして送信）"""
        if destination_name not in self.destinations:
            print(f"無効な目的地名: {destination_name}")
            return False
        
        destination = self.destinations[destination_name]
    
        # 経由点を含む場合とそうでない場合で処理を分ける
        if isinstance(destination, list):
            # 経由点を含む場合は順番に送信
            for point in destination:
                message = json.dumps({
                    "type": "destination",
                    "data": {"name": destination_name, "x": point["x"], "y": point["y"]},
                    "timestamp": time.time()
                })
            
                with self.clients_lock:
                    for client_id, client_info in self.clients.items():
                        try:
                            client_info["socket"].send(message.encode())
                            time.sleep(0.1)  # 送信間隔を設定
                        except Exception as e:
                            print(f"送信エラー: {e}")
                            self.cleanup_client(client_id)
            
                print(f"目的地ポイントを送信: {destination_name} ({point['x']}, {point['y']})")
        else:
            # チェックポイントなど単一座標の場合
            message = json.dumps({
                "type": "destination",
                "data": {"name": destination_name, "x": destination["x"], "y": destination["y"]},
                "timestamp": time.time()
            })
        
            with self.clients_lock:
                for client_id, client_info in self.clients.items():
                    try:
                        client_info["socket"].send(message.encode())
                        print(f"単一目的地を送信: {destination_name}")
                    except Exception as e:
                        print(f"送信エラー: {e}")
                        self.cleanup_client(client_id)
                    
        return True

    def broadcast_route(self, route_name):
        """選択された巡回ルートの目的地を順番に送信"""
        if route_name not in self.routes:
            print(f"無効なルート名: {route_name}")
            return False
            
        route = self.routes[route_name]
        self.current_route = route_name
        
        for destination in route:
            message = json.dumps({
                "type": "destination",
                "data": destination,
                "timestamp": time.time()
            })
            
            with self.clients_lock:
                for client_id, client_info in self.clients.items():
                    try:
                        client_info["socket"].send(message.encode())
                        print(f"目的地を送信: {destination}")
                        time.sleep(0.1)  # 送信間隔を設定
                    except Exception as e:
                        print(f"送信エラー: {e}")
                        self.cleanup_client(client_id)
                        
        return True

    def start(self):
        """サーバーの開始"""
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_socket.bind((self.host, self.port))
        self.server_socket.listen(5)
        print(f"ソケットサーバーを開始しました ({self.host}:{self.port})")
        
        while self.is_running:
            try:
                conn, addr = self.server_socket.accept()
                client_thread = threading.Thread(target=self.handle_client, args=(conn, addr))
                client_thread.daemon = True
                client_thread.start()
            except Exception as e:
                print(f"接続エラー: {e}")

    def get_current_position(self):
        """現在の位置情報を取得"""
        with self.position_lock:
            return dict(self.current_position)  # 複製を返す

# グローバルトラッカーインスタンス
tracker = VehicleTracker()

@app.route('/')
def index():
    """メインページの表示"""
    return render_template('index.html',
                         current_position=tracker.get_current_position(),
                         routes=tracker.routes,
                         destinations=tracker.destinations,
                         current_route=tracker.current_route)

@app.route('/get_position')
def get_position():
    """現在位置の取得"""
    position = tracker.get_current_position()
    if tracker.debug:
        print(f"位置情報をクライアントに送信: {position}")
    return jsonify(position)

@app.route('/set_route/<route_name>')
def set_route(route_name):
    """巡回ルートの設定"""
    if route_name in tracker.routes:
        if tracker.broadcast_route(route_name):
            return jsonify({
                "status": "success",
                "message": f"巡回ルート '{route_name}' を設定しました"
            })
    return jsonify({
        "status": "error",
        "message": "無効なルート名です"
    }), 400

@app.route('/set_destination/<destination_name>')
def set_destination(destination_name):
    """単一の目的地を設定"""
    if destination_name in tracker.destinations:
        if tracker.broadcast_single_destination(destination_name):
            return jsonify({
                "status": "success",
                "message": f"目的地 '{destination_name}' を設定しました"
            })
    return jsonify({
        "status": "error",
        "message": "無効な目的地名です"
    }), 400

if __name__ == "__main__":
    socket_thread = threading.Thread(target=tracker.start)
    socket_thread.daemon = True
    socket_thread.start()
    
    try:
        app.run(host='0.0.0.0', port=8000, debug=True, use_reloader=False)
    except KeyboardInterrupt:
        tracker.is_running = False
        if tracker.server_socket:
            tracker.server_socket.close()
        sys.exit(0)
