#include <micro_ros_arduino.h>
#include <stdio.h>
#include <rcl/rcl.h>
#include <rcl/error_handling.h>
#include <rclc/rclc.h>
#include <rclc/executor.h>
#include <std_msgs/msg/int32.h>
#include <std_msgs/msg/int32_multi_array.h>
#include <nav_msgs/msg/odometry.h>
#include <geometry_msgs/msg/transform_stamped.h>
#include <geometry_msgs/msg/twist.h>
#include <rosidl_runtime_c/string_functions.h>
#include <tf2_msgs/msg/tf_message.h>
#include "Wire.h"
#include "define.h"
#include "quaternion.h"
#include "types.h"

// グローバル変数の定義
volatile int count_l = 0;
volatile int count_r = 0;
int last_count_l = 0;
int last_count_r = 0;
VelocityCommand vel_cmd;
OdometryData odom_data;

unsigned long last_cmd_time = 0.0;
const unsigned long CMD_TIMEOUT = 500;

// グローバル変数の追加
static struct timespec g_last_odom_time;
static bool g_first_odom_calc = true;

// ROS関連の変数
rcl_publisher_t odom_pub;
rcl_publisher_t tf_broadcaster;
rcl_publisher_t enc_pub;
rcl_subscription_t cmd_vel_sub;
rclc_executor_t executor;
rclc_support_t support;
rcl_allocator_t allocator;
rcl_node_t node;
rcl_timer_t timer;
nav_msgs__msg__Odometry odom_msg;
geometry_msgs__msg__TransformStamped odom_tf;
std_msgs__msg__Int32MultiArray enc_msg;
geometry_msgs__msg__Twist cmd_vel_msg;

rcl_publisher_t debug_pub;
std_msgs__msg__Int32MultiArray debug_msg;

void cmd_vel_callback(const void * msgin) {
    const geometry_msgs__msg__Twist * msg = (const geometry_msgs__msg__Twist *)msgin;
    vel_cmd.linear_x = msg->linear.x;
    vel_cmd.angular_z = msg->angular.z;
    
    // タイムアウトチェック用に時間を更新
    last_cmd_time = millis();
    
    // 速度コマンドの処理
    process_velocity_command(vel_cmd);
}

// calculate_odometry()の修正版
void calculate_odometry() {
    struct timespec current_time;
    clock_gettime(CLOCK_REALTIME, &current_time);
    
    if (g_first_odom_calc) {
        g_last_odom_time = current_time;
        g_first_odom_calc = false;
        return;
    }
    
    double dt = (current_time.tv_sec - g_last_odom_time.tv_sec) +
                (current_time.tv_nsec - g_last_odom_time.tv_nsec) / 1e9;
    
    if (dt <= 0.0 || dt > 1.0) {
        g_last_odom_time = current_time;
        return;
    }

    // エンコーダ値の読み取り
    int current_count_l, current_count_r;
    encoder_get(&current_count_l, &current_count_r);
    
    // エンコーダの差分を計算
    int delta_left = current_count_l - last_count_l;
    int delta_right = -(current_count_r - last_count_r);  // 右輪は反転

    // エンコーダの差分が閾値以下の場合は移動なしと判断
    const int ENCODER_NOISE_THRESHOLD = 0;  // ノイズとみなす閾値
    if (abs(delta_left) <= ENCODER_NOISE_THRESHOLD && 
        abs(delta_right) <= ENCODER_NOISE_THRESHOLD) {
        // 速度をゼロに設定
        wheel_velocity_left = 0.0;
        wheel_velocity_right = 0.0;
        odom_data.velocity = 0.0;
        odom_data.omega = 0.0;
        
        // エンコーダの現在値を保存
        last_count_l = current_count_l;
        last_count_r = current_count_r;
        g_last_odom_time = current_time;
        return;
    }
    
    // motor_controllerで計算された実際の速度を使用
    double v_left = wheel_velocity_left;
    double v_right = wheel_velocity_right;
    
    // 速度が非常に小さい場合は無視
    const double MIN_VELOCITY = 0.0001;  // m/s
    if (fabs(v_left) < MIN_VELOCITY && fabs(v_right) < MIN_VELOCITY) {
        v_left = 0.0;
        v_right = 0.0;
    }
    
    // 台車全体の並進速度と角速度を計算
    odom_data.velocity = (v_right + v_left) / 2.0;  // 並進速度 [m/s]
    odom_data.omega = (v_left - v_right) / WHEEL_BASE;  // 角速度 [rad/s]
    
    // 移動距離と回転角の計算
    double delta_distance = odom_data.velocity * dt;  // 並進移動距離 [m]
    double delta_theta = odom_data.omega * dt;  // 回転角 [rad]
    
    // 有意な移動がある場合のみ位置を更新
    if (fabs(delta_distance) > 0.0001 || fabs(delta_theta) > 0.0001) {
        // 姿勢角の更新
        odom_data.theta += delta_theta;
        
        // 姿勢角の正規化（-π から π の範囲に収める）
        while (odom_data.theta > PI) odom_data.theta -= 2.0 * PI;
        while (odom_data.theta < -PI) odom_data.theta += 2.0 * PI;
        
        // 位置の更新
        double cos_theta = cos(odom_data.theta);
        double sin_theta = sin(odom_data.theta);
        odom_data.x += delta_distance * cos_theta;
        odom_data.y += delta_distance * sin_theta;
    }
    
    // エンコーダの現在値を保存
    last_count_l = current_count_l;
    last_count_r = current_count_r;
    
    // デバッグ情報の更新
    debug_msg.data.data[0] = delta_left;
    debug_msg.data.data[1] = delta_right;
    debug_msg.data.data[2] = (int32_t)(dt * 1000000.0);  // マイクロ秒単位
    debug_msg.data.data[3] = (int32_t)(v_left * 1000);   // mm/s
    debug_msg.data.data[4] = (int32_t)(v_right * 1000);  // mm/s
    debug_msg.data.data[5] = current_count_l;  // 生のカウント値
    debug_msg.data.data[6] = current_count_r;
    debug_msg.data.data[7] = (int32_t)(odom_data.velocity * 1000);  // mm/s
    debug_msg.data.data[8] = (int32_t)(odom_data.omega * 1000);     // mrad/s
    
    rcl_publish(&debug_pub, &debug_msg, NULL);
    g_last_odom_time = current_time;
}

void timer_callback(rcl_timer_t * timer, int64_t last_call_time) {
    RCLC_UNUSED(last_call_time);
    if (timer != NULL) {
        // オドメトリの計算
        calculate_odometry();
        
        // タイムスタンプの更新
        struct timespec ts;
        clock_gettime(CLOCK_REALTIME, &ts);
        odom_msg.header.stamp.sec = ts.tv_sec;
        odom_msg.header.stamp.nanosec = ts.tv_nsec;
        odom_tf.header.stamp = odom_msg.header.stamp;
        
        // オドメトリメッセージの更新
        odom_msg.pose.pose.position.x = odom_data.x;
        odom_msg.pose.pose.position.y = odom_data.y;
        odom_msg.pose.pose.position.z = 0.0;
        
        // クォータニオンの計算と設定
        Quaternion q = toQuaternion(odom_data.theta, 0, 0);
        odom_msg.pose.pose.orientation.x = q.x;
        odom_msg.pose.pose.orientation.y = q.y;
        odom_msg.pose.pose.orientation.z = q.z;
        odom_msg.pose.pose.orientation.w = q.w;
        
        // 速度の設定
        odom_msg.twist.twist.linear.x = odom_data.velocity;
        odom_msg.twist.twist.angular.z = odom_data.omega;
        
        // TF変換の更新
        odom_tf.transform.translation.x = odom_data.x;
        odom_tf.transform.translation.y = odom_data.y;
        odom_tf.transform.translation.z = 0.0;
        odom_tf.transform.rotation = odom_msg.pose.pose.orientation;
        
        // エンコーダ値の公開
        enc_msg.data.data[0] = count_l;
        enc_msg.data.data[1] = count_r;
        
        // メッセージのパブリッシュ
        rcl_publish(&enc_pub, &enc_msg, NULL);
        rcl_publish(&odom_pub, &odom_msg, NULL);
        rcl_publish(&tf_broadcaster, &odom_tf, NULL);
    }
}
void setup() {
    Serial.begin(115200);
    encoder_open();
    set_microros_transports();
    motor_init();

    // オドメトリデータの初期化（初期位置）
    odom_data.x = 50.5;     
    odom_data.y = 5.5;      
    odom_data.theta = - (PI / 2.0); 
    odom_data.velocity = 0.0;   // 明示的に速度を初期化
    odom_data.omega = 0.0;      // 明示的に角速度を初期化

    // エンコーダのリセットと初期値の読み取り
    encoder_reset();
    encoder_get(&last_count_l, &last_count_r);  // 安全な初期値の読み取り
    
    // オドメトリ計算用の時間初期化
    clock_gettime(CLOCK_REALTIME, &g_last_odom_time);
    g_first_odom_calc = true;
    
    delay(1000);
    
    allocator = rcl_get_default_allocator();
    rclc_support_init(&support, 0, NULL, &allocator);
    rclc_node_init_default(&node, "ESP32_odom_node", "", &support);
    
    // パブリッシャーの初期化
    rclc_publisher_init_default(
        &odom_pub,
        &node,
        ROSIDL_GET_MSG_TYPE_SUPPORT(nav_msgs, msg, Odometry),
        "/odom"
    );
    
    rclc_publisher_init_default(
        &tf_broadcaster,
        &node,
        ROSIDL_GET_MSG_TYPE_SUPPORT(geometry_msgs, msg, TransformStamped),
        "/tf"
    );
    
    rclc_publisher_init_default(
        &enc_pub,
        &node,
        ROSIDL_GET_MSG_TYPE_SUPPORT(std_msgs, msg, Int32MultiArray),
        "/encoder"
    );
    
    // タイマー周期
    const unsigned int timer_timeout = 50;  //50ms
    rclc_timer_init_default(
        &timer,
        &support,
        RCL_MS_TO_NS(timer_timeout),
        timer_callback
    );
    
    rclc_subscription_init_default(
        &cmd_vel_sub,
        &node,
        ROSIDL_GET_MSG_TYPE_SUPPORT(geometry_msgs, msg, Twist),
        "cmd_vel"
    );

    // デバッグパブリッシャーの初期化
    rclc_publisher_init_default(
        &debug_pub,
        &node,
        ROSIDL_GET_MSG_TYPE_SUPPORT(std_msgs, msg, Int32MultiArray),
        "/odom_debug"
    );

    // デバッグメッセージの初期化
    debug_msg.data.size = 9;  // delta_left, delta_right, dt_micros, METERS_PER_COUNT_scaled
    debug_msg.data.data = (int32_t *)malloc(debug_msg.data.size * sizeof(int32_t));
    
    // フレームIDの設定
    rosidl_runtime_c__String__assign(&odom_msg.header.frame_id, "odom");
    rosidl_runtime_c__String__assign(&odom_msg.child_frame_id, "base_link");
    rosidl_runtime_c__String__assign(&odom_tf.header.frame_id, "odom");
    rosidl_runtime_c__String__assign(&odom_tf.child_frame_id, "base_link");
    
    // エンコーダメッセージの初期化
    enc_msg.data.size = 2;
    enc_msg.data.data = (int32_t *)malloc(enc_msg.data.size * sizeof(int32_t));
    enc_msg.data.data[0] = 0;
    enc_msg.data.data[1] = 0;
    
    // エグゼキューターの設定
    rclc_executor_init(&executor, &support.context, 2, &allocator);
    rclc_executor_add_timer(&executor, &timer);
    rclc_executor_add_subscription(&executor, &cmd_vel_sub, &cmd_vel_msg, &cmd_vel_callback, ON_NEW_DATA);
}

void loop() {
    unsigned long current_time = millis();
    
    // cmd_velのタイムアウトチェック
    if (current_time - last_cmd_time > CMD_TIMEOUT) {
        vel_cmd.linear_x = 0;
        vel_cmd.angular_z = 0;
        emergency_stop();  // 緊急停止処理を呼び出し
    } else {
        update_velocity_control();  // 通常の速度制御更新
    }
    
    rclc_executor_spin_some(&executor, RCL_MS_TO_NS(10));
}
