// motor_controller.ino
#include "define.h"
#include "types.h"

// グローバル変数の定義
double wheel_velocity_left = 0;
double wheel_velocity_right = 0;
static unsigned long last_time = 0;  // last_time変数をグローバルで定義
const double MAX_ROTATION_SPEED = 1.0;  // 最大回転速度 [m/s]

// モータパラメータの初期化
MotorParameters left_motor_params = {
    KP_MOTOR_L, KI_MOTOR_L, KD_MOTOR_L,
    MIN_POWER_L, MAX_POWER_L,
    LEFT_MOTOR_POWER_SCALE,
    true  // 左モータは方向反転
};

MotorParameters right_motor_params = {
    KP_MOTOR_R, KI_MOTOR_R, KD_MOTOR_R,
    MIN_POWER_R, MAX_POWER_R,
    RIGHT_MOTOR_POWER_SCALE,
    false // 右モータは方向そのまま
};

// モータコントローラの初期化
MotorController left_motor(left_motor_params);
MotorController right_motor(right_motor_params);

// MotorControllerクラスの実装
MotorController::MotorController(const MotorParameters& parameters) 
    : target(0), previous_error(0), integral(0), filtered_derivative(0),
      last_output(0), last_velocity(0), is_rotation_mode(false),
      params(parameters) {}

void MotorController::setTarget(double new_target) {
    if (fabs(target - new_target) > 0.001) {
        target = new_target;
        if (fabs(new_target) < 0.001) {
            reset();
        }
    }
}

void MotorController::setRotationMode(bool rotation_mode) {
    if (is_rotation_mode != rotation_mode) {
        is_rotation_mode = rotation_mode;
        reset();
    }
}

void MotorController::reset() {
    previous_error = 0;
    integral = 0;
    filtered_derivative = 0;
    last_output = 0;
    last_velocity = 0;
}

double MotorController::compute(double current_value, double dt) {
    return is_rotation_mode ? 
           computeRotation(current_value, dt) : 
           computeLinear(current_value, dt);
}

double MotorController::computeLinear(double current_value, double dt) {
    double error = target - current_value;
    
    // 速度変化の検出と処理
    double velocity_change = (current_value - last_velocity) / dt;
    last_velocity = current_value;
    
    // 積分項の処理
    if (fabs(velocity_change) > 0.1) {
        integral = 0;
    } else {
        if (fabs(error) < 0.15) {
            double error_ratio = fabs(error) / 0.15;
            double integral_scale = (1.0 - error_ratio) * 0.4;
            integral += error * dt * integral_scale;
        }
        integral *= (1.0 - dt * 0.2);
    }
    
    // 積分制限
    double integral_limit = params.max_power / 30;
    integral = fmax(fmin(integral, integral_limit), -integral_limit);
    
    // 微分項の計算と平滑化
    double derivative = (error - previous_error) / dt;
    filtered_derivative = 0.97 * filtered_derivative + 0.03 * derivative;
    previous_error = error;
    
    // PID出力の計算
    double output = params.kp * error + 
                   params.ki * integral + 
                   params.kd * filtered_derivative;
    
    // デッドバンド処理
    if (fabs(output) < params.min_power) {
        if (fabs(target) > 0.001) {
            if (fabs(last_output) > params.min_power) {
                output = last_output * 0.95;
            } else {
                double start_output = target > 0 ? params.min_power : -params.min_power;
                output = last_output * 0.9 + start_output * 0.1;
            }
        } else {
            output = 0;
        }
    }
    
    last_output = output;
    return fmax(fmin(output * params.power_scale, params.max_power), 
               -params.max_power);
}

double MotorController::computeRotation(double current_value, double dt) {
    // 速度誤差の計算
    double error = target - current_value;
    
    // PID制御の計算
    // 速度変化の検出と処理
    double velocity_change = (current_value - last_velocity) / dt;
    last_velocity = current_value;
    
    // 積分項の処理（回転モード用に調整）
    if (fabs(velocity_change) > 0.1) {
        integral = 0;
    } else {
        if (fabs(error) < 0.15) {
            double error_ratio = fabs(error) / 0.15;
            double integral_scale = (1.0 - error_ratio) * 0.3; // 回転モード用に調整
            integral += error * dt * integral_scale;
        }
        integral *= (1.0 - dt * 0.2);
    }
    
    // 積分制限
    double integral_limit = params.max_power / 25; // 回転モード用に調整
    integral = fmax(fmin(integral, integral_limit), -integral_limit);
    
    // 微分項の計算と平滑化
    double derivative = (error - previous_error) / dt;
    filtered_derivative = 0.97 * filtered_derivative + 0.03 * derivative;
    previous_error = error;
    
    // PID出力の計算
    double output = params.kp * error + 
                   params.ki * integral + 
                   params.kd * filtered_derivative;
    
    // 最小出力の保証とスムージング
    if (fabs(output) < params.min_power) {
        if (fabs(target) > 0.001) {
            if (fabs(last_output) > params.min_power) {
                output = last_output * 0.95;
            } else {
                double start_output = target > 0 ? params.min_power : -params.min_power;
                output = last_output * 0.9 + start_output * 0.1;
            }
        } else {
            output = 0;
        }
    }
    
    last_output = output;
    return fmax(fmin(output * params.power_scale, params.max_power), 
               -params.max_power);
}

// モータ初期化関数
void motor_init() {
    ledcSetup(0, PWM_FREQ, PWM_RESOLUTION);
    ledcSetup(1, PWM_FREQ, PWM_RESOLUTION);
    
    ledcAttachPin(PIN_PWM_L, 0);
    ledcAttachPin(PIN_PWM_R, 1);
    
    pinMode(PIN_DIR_L, OUTPUT);
    pinMode(PIN_DIR_R, OUTPUT);
    
    last_time = micros();
}

// モータ出力設定関数
void set_motor_power(int motor_pin, int dir_pin, double power, bool reverse_direction) {
    bool direction = reverse_direction ? (power < 0) : (power >= 0);
    double abs_power = abs(power);
    
    // PWM値の制限
    int pwm = (int)abs_power;
    pwm = fmin(pwm, PWM_MAX);
    
    digitalWrite(dir_pin, direction ? HIGH : LOW);
    ledcWrite(motor_pin == PIN_PWM_L ? 0 : 1, pwm);
}

// 速度制御更新関数
void update_velocity_control() {
    static unsigned long last_time = 0;
    unsigned long current_time = micros();
    double dt = (current_time - last_time) / 1000000.0;
    
    if (dt < 0.001) return;
    
    // エンコーダ値の変化量計算
    int delta_left = count_l - last_count_l;
    int delta_right = -(count_r - last_count_r);
    
    // 速度計算（移動平均フィルタ付き）
    const int FILTER_SIZE = 2;
    static double vel_left_array[FILTER_SIZE] = {0};
    static double vel_right_array[FILTER_SIZE] = {0};
    static int filter_index = 0;
    
    vel_left_array[filter_index] = (delta_left / COUNTS_PER_REV) * 
                                  2.0 * PI * WHEEL_RADIUS / dt;
    vel_right_array[filter_index] = (delta_right / COUNTS_PER_REV) * 
                                   2.0 * PI * WHEEL_RADIUS / dt;
    
    double velocity_left = 0;
    double velocity_right = 0;
    for(int i = 0; i < FILTER_SIZE; i++) {
        velocity_left += vel_left_array[i];
        velocity_right += vel_right_array[i];
    }
    velocity_left /= FILTER_SIZE;
    velocity_right /= FILTER_SIZE;
    
    // グローバル変数の更新
    wheel_velocity_left = velocity_left;
    wheel_velocity_right = velocity_right;
    
    filter_index = (filter_index + 1) % FILTER_SIZE;
    
    // PID制御による出力計算
    double power_left = left_motor.compute(velocity_left, dt);
    double power_right = right_motor.compute(velocity_right, dt);
    
    // モータ出力の設定
    set_motor_power(PIN_PWM_L, PIN_DIR_L, power_left, left_motor_params.reverse_direction);
    set_motor_power(PIN_PWM_R, PIN_DIR_R, power_right, right_motor_params.reverse_direction);
    
    // エンコーダカウントの更新
    last_count_l = count_l;
    last_count_r = count_r;
    last_time = current_time;
}

// 目標速度設定関数
void set_motor_velocities(double left_velocity, double right_velocity, bool rotation_mode) {
    left_motor.setTarget(left_velocity);
    right_motor.setTarget(right_velocity);
    
    left_motor.setRotationMode(rotation_mode);
    right_motor.setRotationMode(rotation_mode);
}

// cmd_vel コマンドを処理する関数
void process_velocity_command(const VelocityCommand& cmd) {
    if (fabs(cmd.linear_x) > 0.001 && fabs(cmd.angular_z) < 0.001) {
        // 直進モード
        set_motor_velocities(cmd.linear_x, cmd.linear_x, false);
    } 
    else if (fabs(cmd.linear_x) < 0.001 && fabs(cmd.angular_z) > 0.001) {
        // 回転モード
        const double MAX_ROTATION_SPEED = 0.25;  // 最大回転速度 [m/s]
        const double MIN_ROTATION_SPEED = 0.05;  // 最小回転速度 [m/s]
        
        // angular_zの絶対値を0-1の範囲に正規化（典型的な値範囲を2.0 rad/sとして）
        double normalized_omega = fmin(fabs(cmd.angular_z) / 2.0, 1.0);
        
        // 回転速度を非線形にスケーリング（より細かい制御を可能に）
        double rotation_speed = MIN_ROTATION_SPEED + 
            (MAX_ROTATION_SPEED - MIN_ROTATION_SPEED) * (normalized_omega * normalized_omega);
            
        if (cmd.angular_z > 0) {  // 反時計回り
            set_motor_velocities(rotation_speed, -rotation_speed, true);
        } else {  // 時計回り
            set_motor_velocities(-rotation_speed, rotation_speed, true);
        }
    }
    else if (fabs(cmd.linear_x) > 0.001 && fabs(cmd.angular_z) > 0.001) {
        // 旋回モード（直進と回転の組み合わせ）
        double v_rotation = (cmd.angular_z * WHEEL_BASE) / 2.0;
        double v_left = cmd.linear_x - v_rotation;
        double v_right = cmd.linear_x + v_rotation;
        
        // 速度の正規化（最大速度を超えないように調整）
        double max_speed = fmax(fabs(v_left), fabs(v_right));
        if (max_speed > MAX_ROTATION_SPEED) {
            double scale = MAX_ROTATION_SPEED / max_speed;
            v_left *= scale;
            v_right *= scale;
        }
        
        set_motor_velocities(v_left, v_right, false);
    }
    else {
        // 停止
        set_motor_velocities(0, 0, false);
    }
}

// モータの緊急停止
void emergency_stop() {
    set_motor_velocities(0, 0, false);
    left_motor.reset();
    right_motor.reset();
    set_motor_power(PIN_PWM_L, PIN_DIR_L, 0, left_motor_params.reverse_direction);
    set_motor_power(PIN_PWM_R, PIN_DIR_R, 0, right_motor_params.reverse_direction);
}
