# === robot_control/path_planner.py ===
from typing import List, Tuple
import numpy as np
from geometry_msgs.msg import Pose, PoseStamped
from nav_msgs.msg import Path

class PathPlanner:
    def __init__(self):
        self.resolution = 0.065  # マップの解像度（m/セル）
        self.robot_radius = 0.4  # ロボットの半径（m）
        
    def create_path(self, start_pose: Pose, goal_pose: Pose,
                   obstacle_positions: List[Tuple[float, float]]) -> Path:
        """
        A*アルゴリズムを使用してパスを生成
        """
        # パスの生成（簡易実装）
        path = Path()
        path.header.frame_id = "map"
        
        # 直線的な中間点を生成
        num_points = 5
        for i in range(num_points + 1):
            t = i / num_points
            x = start_pose.position.x * (1-t) + goal_pose.position.x * t
            y = start_pose.position.y * (1-t) + goal_pose.position.y * t
            
            pose = PoseStamped()
            pose.header.frame_id = "map"
            pose.pose.position.x = x
            pose.pose.position.y = y
            
            # 簡単な障害物回避
            for obs_x, obs_y in obstacle_positions:
                if np.sqrt((x - obs_x)**2 + (y - obs_y)**2) < self.robot_radius:
                    # 障害物を回避する位置に調整
                    angle = np.arctan2(y - obs_y, x - obs_x)
                    pose.pose.position.x = obs_x + (self.robot_radius + 0.1) * np.cos(angle)
                    pose.pose.position.y = obs_y + (self.robot_radius + 0.1) * np.sin(angle)
            
            path.poses.append(pose)
        
        return path

    def optimize_path(self, path: Path) -> Path:
        if not path or not path.poses:
            return path
    
        optimized_path = Path()
        optimized_path.header = path.header
        optimized_path.poses = path.poses  # 既存のポーズを保持
    
        return optimized_path
        
