#!/usr/bin/env python3

import rclpy
from rclpy.node import Node
from nav_msgs.msg import Odometry
from geometry_msgs.msg import Twist, PoseStamped, Pose
from std_msgs.msg import Int32MultiArray
from sensor_msgs.msg import LaserScan  # LiDARデータ用
from robot_interfaces.msg import RobotStatus, EmergencyStop, NavigationGoal
from nav2_msgs.action import NavigateToPose
from rclpy.action import ActionClient
import numpy as np
import time

class RobotNavigationControl(Node):
    def __init__(self):
        super().__init__('robot_navigation_control')
        
        self.declare_parameters(
            namespace='',
            parameters=[
                ('max_linear_velocity', 0.5),
                ('max_angular_velocity', 5.0),
                ('position_tolerance', 0.1),
                ('orientation_tolerance', 0.1),
                ('safety_distance', 0.8),         # 障害物との安全距離を0.8mに増加
                ('emergency_distance', 0.5),      # 緊急停止距離 0.5m
                ('scan_angle_min', 45.0),       # front スキャン
                ('scan_angle_max', 135.0),
                ('pole_filter.distance_min', 0.20),  # 支柱判定の最小距離
                ('pole_filter.distance_max', 0.50),  # 支柱判定の最大距離
                ('pole_filter.angle_range', 20.0),   # 支柱判定の角度範囲（度）
                ('valid_range.min', 0.45),          # 有効な最小検出距離
                ('valid_range.max', 8.0)   
            ]
        )
        
        self.max_linear_velocity = self.get_parameter('max_linear_velocity').value
        self.max_angular_velocity = self.get_parameter('max_angular_velocity').value
        self.position_tolerance = self.get_parameter('position_tolerance').value
        self.orientation_tolerance = self.get_parameter('orientation_tolerance').value
        self.safety_distance = self.get_parameter('safety_distance').value
        self.emergency_distance = self.get_parameter('emergency_distance').value
        self.scan_angle_min = np.radians(self.get_parameter('scan_angle_min').value)
        self.scan_angle_max = np.radians(self.get_parameter('scan_angle_max').value)
        
        self.cmd_vel_pub = self.create_publisher(Twist, 'cmd_vel', 10)
        self.robot_status_pub = self.create_publisher(RobotStatus, 'robot_status', 10)
        self.initial_pose_pub = self.create_publisher(PoseStamped, 'initialpose', 10)
        
        self.odom_sub = self.create_subscription(
            Odometry, 
            'odom', 
            self.odom_callback, 
            10
        )
        self.encoder_sub = self.create_subscription(
            Int32MultiArray, 
            'encoder', 
            self.encoder_callback, 
            10
        )
        self.emergency_sub = self.create_subscription(
            EmergencyStop, 
            'emergency_stop', 
            self.emergency_callback, 
            10
        )
        self.goal_sub = self.create_subscription(
            NavigationGoal, 
            'navigation_goal', 
            self.goal_callback, 
            10
        )
        self.scan_sub = self.create_subscription(
            LaserScan,
            'scan',
            self.scan_callback,
            10
        )
            
        self.nav_client = ActionClient(self, NavigateToPose, 'navigate_to_pose')
        
        self.current_pose = None
        self.current_velocity = None
        self.emergency_stop = False
        self.navigation_active = False
        self.goal_pose = None
        
        self.obstacle_detected = False
        self.min_distance = float('inf')
        self.scan_data = None
        
        self.encoder_left = 0
        self.encoder_right = 0
        
        # 初期位置の設定
        initial_pose = PoseStamped()
        initial_pose.header.frame_id = 'map'
        initial_pose.header.stamp = self.get_clock().now().to_msg()
        initial_pose.pose.position.x = 50.5  # クリエイティブラボのX座標
        initial_pose.pose.position.y = 5.5   # クリエイティブラボのY座標
        initial_pose.pose.orientation.z = -0.707107 # sin(-π/4) (南向き)
        initial_pose.pose.orientation.w = 0.707107  # cos(-π/4) (南向き)
        
        # システムが安定するまで少し待機
        time.sleep(1.0)
        
        self.initial_pose_pub.publish(initial_pose)
        
        self.create_timer(0.1, self.control_loop)
        
        self.get_logger().info('Robot Navigation Control node initialized')
    
    def odom_callback(self, msg):
        self.current_pose = msg.pose.pose
        self.current_velocity = msg.twist.twist
        
    def encoder_callback(self, msg):
        self.encoder_left = msg.data[0]
        self.encoder_right = msg.data[1]
      
    def scan_callback(self, msg):
        """LiDARデータの処理（ドライバーでフィルタリング済み）"""
        try:
            # 前方のデータのみを使用
            start_idx = int((np.radians(45) - msg.angle_min) / msg.angle_increment)
            end_idx = int((np.radians(135) - msg.angle_min) / msg.angle_increment)
        
            # 配列の範囲チェック
            if start_idx < 0:
                start_idx = 0
            if end_idx > len(msg.ranges):
                end_idx = len(msg.ranges)
            
            # 前方範囲のデータを抽出
            front_ranges = []
            for i in range(start_idx, end_idx):
                r = msg.ranges[i]
                if np.isfinite(r) and msg.range_min <= r <= msg.range_max:
                    front_ranges.append(r)
        
            # 前方の最小距離を計算
            if front_ranges:
                self.min_distance = min(front_ranges)
            
                # 障害物検出と緊急停止の判定
                self.obstacle_detected = self.min_distance < self.safety_distance
            
                if self.min_distance < self.emergency_distance and not self.emergency_stop:
                    self.emergency_stop = True
                    self.stop_robot()
                    self.get_logger().warn(
                        f'Emergency stop: Obstacle detected at {self.min_distance:.2f}m'
                    )
                elif self.min_distance >= self.emergency_distance:
                    self.emergency_stop = False
            else:
                # 前方に有効なデータがない場合
                self.min_distance = float('inf')
                self.obstacle_detected = False
                self.emergency_stop = False
            
            # スキャンデータを保存
            self.scan_data = msg
        
        except Exception as e:
            self.get_logger().error(f'Error in scan_callback: {str(e)}')

    def _normalize_angle(self, angle):
        """角度を-πからπの範囲に正規化"""
        while angle > np.pi:
            angle -= 2 * np.pi
        while angle < -np.pi:
            angle += 2 * np.pi
        return angle
  
    def emergency_callback(self, msg):
        self.emergency_stop = msg.stop
        if self.emergency_stop:
            self.stop_robot()
            self.get_logger().warn('Emergency stop activated!')
        
    def _transform_coordinates(self, pose: Pose) -> Pose:
        """
        座標系の変換
        Webサーバーの座標系からROSの座標系への変換
        """
        transformed = Pose()
        # Web座標系からROSの座標系への変換
        transformed.position.x = pose.position.x
        transformed.position.y = pose.position.y
        transformed.orientation = pose.orientation
        return transformed

    def goal_callback(self, msg):
        if self.emergency_stop:
            self.get_logger().warn('Cannot start navigation while emergency stop is active')
            return
            
        # 座標変換を適用
        transformed_goal = self._transform_coordinates(msg.pose)
        msg.pose = transformed_goal
        self.goal_pose = msg
        self.start_navigation()
        
    def start_navigation(self):
        if not self.nav_client.wait_for_server(timeout_sec=1.0):
            self.get_logger().error('Navigation action server not available')
            return
            
        goal_msg = NavigateToPose.Goal()
        goal_msg.pose.header.frame_id = 'map'
        goal_msg.pose.header.stamp = self.get_clock().now().to_msg()
        goal_msg.pose.pose = self.goal_pose.pose
        
        self.navigation_active = True
        self.nav_client.send_goal_async(goal_msg, feedback_callback=self.navigation_feedback_callback)
            
    def navigation_feedback_callback(self, feedback_msg):
        feedback = feedback_msg.feedback
        status_msg = RobotStatus()
        status_msg.navigation_active = self.navigation_active
        status_msg.current_pose = self.current_pose
        status_msg.distance_remaining = feedback.distance_remaining
        self.robot_status_pub.publish(status_msg)
        
    def stop_robot(self):
        stop_msg = Twist()
        stop_msg.linear.x = 0.0
        stop_msg.angular.z = 0.0
        self.cmd_vel_pub.publish(stop_msg)
        self.navigation_active = False
        
    def velocity_controller(self):
        """速度制御と衝突回避の実装（改良版）"""
        cmd_vel = Twist()

        if not self.scan_data or not self.current_pose or not self.goal_pose:
            return cmd_vel

        # 目標地点までの相対位置を計算
        dx = self.goal_pose.position.x - self.current_pose.position.x
        dy = self.goal_pose.position.y - self.current_pose.position.y

        # 目標までの距離と角度を計算
        distance = np.sqrt(dx*dx + dy*dy)
        target_angle = np.arctan2(dy, dx)

        # 現在の向きを取得
        current_yaw = self._get_yaw_from_quaternion(self.current_pose.orientation)

        # 角度差の計算（-πからπの範囲に正規化）
        angle_diff = np.arctan2(np.sin(target_angle - current_yaw), 
                               np.cos(target_angle - current_yaw))

        # パラメータの取得
        slowdown_distance = self.get_parameter('velocity_control.slowdown_distance').value
        min_linear_velocity = self.get_parameter('velocity_control.min_linear_velocity').value
        rotation_scaling = self.get_parameter('velocity_control.rotation_scaling').value

        # 速度制御のロジック
        if abs(angle_diff) > 0.5:  # ≒30度以上の場合
            # 方向修正を優先
            cmd_vel.angular.z = rotation_scaling * angle_diff
            cmd_vel.linear.x = min_linear_velocity
        else:
             # 直進速度の計算（距離に応じた減速）
            if distance > slowdown_distance:
                cmd_vel.linear.x = self.max_linear_velocity
            else:
                # 残り距離に応じて線形に減速
                speed_ratio = distance / slowdown_distance
                cmd_vel.linear.x = max(
                    min_linear_velocity,
                    speed_ratio * self.max_linear_velocity
                )
        
            # 方向修正の角速度
            cmd_vel.angular.z = rotation_scaling * angle_diff

        # 障害物が検出された場合の速度調整
        if self.obstacle_detected:
            slow_factor = max(0.3, self.min_distance / self.safety_distance)
            cmd_vel.linear.x *= slow_factor
            cmd_vel.angular.z *= slow_factor

        # 速度の制限
        cmd_vel.linear.x = max(-self.max_linear_velocity, 
                              min(self.max_linear_velocity, cmd_vel.linear.x))
        cmd_vel.angular.z = max(-self.max_angular_velocity,
                               min(self.max_angular_velocity, cmd_vel.angular.z))

        return cmd_vel

    def _get_yaw_from_quaternion(self, q):
        """クォータニオンからヨー角を取得"""
        # クォータニオンから回転行列を計算
        sinr_cosp = 2.0 * (q.w * q.z + q.x * q.y)
        cosr_cosp = 1.0 - 2.0 * (q.y * q.y + q.z * q.z)
        return np.arctan2(sinr_cosp, cosr_cosp)

    def _euler_from_quaternion(self, x, y, z, w):
        """クォータニオンからオイラー角への変換"""
        # ロール（x軸周りの回転）
        sinr_cosp = 2 * (w * x + y * z)
        cosr_cosp = 1 - 2 * (x * x + y * y)
        roll = np.arctan2(sinr_cosp, cosr_cosp)
    
        # ピッチ（y軸周りの回転）
        sinp = 2 * (w * y - z * x)
        if abs(sinp) >= 1:
            pitch = np.copysign(np.pi / 2, sinp)
        else:
            pitch = np.arcsin(sinp)
    
        # ヨー（z軸周りの回転）
        siny_cosp = 2 * (w * z + x * y)
        cosy_cosp = 1 - 2 * (y * y + z * z)
        yaw = np.arctan2(siny_cosp, cosy_cosp)
    
        return roll, pitch, yaw

    def control_loop(self):
        if self.emergency_stop:
            self.stop_robot()
            return
                
        if not self.navigation_active:
            return
                
        # 速度制御の実行
        cmd_vel = self.velocity_controller()
        self.cmd_vel_pub.publish(cmd_vel)
        
        # ステータス更新
        status_msg = RobotStatus()
        status_msg.navigation_active = self.navigation_active
        status_msg.current_pose = self.current_pose
        status_msg.emergency_stop = self.emergency_stop
        status_msg.obstacle_detected = self.obstacle_detected
        status_msg.min_distance = self.min_distance if self.min_distance != float('inf') else 0.0
        self.robot_status_pub.publish(status_msg)

def main(args=None):
    rclpy.init(args=args)
    node = RobotNavigationControl()
    try:
        rclpy.spin(node)
    except KeyboardInterrupt:
        pass
    finally:
        node.destroy_node()
        rclpy.shutdown()

if __name__ == '__main__':
    main()
