# === robot_control/robot_utilities.py ===
import math
import numpy as np
from geometry_msgs.msg import Pose, Point, Quaternion
from transforms3d.euler import euler2quat, quat2euler

class RobotUtilities:
    @staticmethod
    def calculate_distance(pose1: Pose, pose2: Pose) -> float:
        """2つのポーズ間の距離を計算"""
        return math.sqrt(
            (pose1.position.x - pose2.position.x) ** 2 +
            (pose1.position.y - pose2.position.y) ** 2
        )

    @staticmethod
    def calculate_angle_difference(pose1: Pose, pose2: Pose) -> float:
        """2つのポーズ間の角度差を計算"""
        q1 = [pose1.orientation.x, pose1.orientation.y,
              pose1.orientation.z, pose1.orientation.w]
        q2 = [pose2.orientation.x, pose2.orientation.y,
              pose2.orientation.z, pose2.orientation.w]
        
        yaw1 = quat2euler(q1)[2]
        yaw2 = quat2euler(q2)[2]
        
        diff = yaw2 - yaw1
        return math.atan2(math.sin(diff), math.cos(diff))

    @staticmethod
    def check_safety_zone(scan_data, safety_params):
        """
        フィルタリング済みのLiDARデータから安全領域をチェック
    
        Args:
            scan_data: センサーデータ (LaserScan)
            safety_params: 安全パラメータ (dict)
                - emergency_distance: 緊急停止距離
                - safety_distance: 安全距離
                - angle_min: 最小スキャン角度
                - angle_max: 最大スキャン角度
    
        Returns:
            bool: 安全かどうか
            float: 最も近い障害物までの距離
            float: その障害物の方向（ラジアン）
        """
        if scan_data is None:
            return True, float('inf'), 0.0

        # 前方範囲の設定
        start_idx = int((safety_params.get('angle_min', -0.785) - scan_data.angle_min) 
                        / scan_data.angle_increment)
        end_idx = int((safety_params.get('angle_max', 0.785) - scan_data.angle_min) 
                      / scan_data.angle_increment)

        # 配列の範囲チェック
        start_idx = max(0, start_idx)
        end_idx = min(len(scan_data.ranges), end_idx)

        min_dist = float('inf')
        min_angle = 0.0
        is_safe = True

        # 前方範囲のみチェック
        for i in range(start_idx, end_idx):
            r = scan_data.ranges[i]
            if scan_data.range_min <= r <= scan_data.range_max:
                angle = scan_data.angle_min + i * scan_data.angle_increment
                if r < safety_params['emergency_distance']:
                    is_safe = False
                if r < min_dist:
                    min_dist = r
                    min_angle = angle

        return is_safe, min_dist, min_angle

    @staticmethod
    def calculate_avoidance_velocity(current_velocity, obstacle_distance, obstacle_angle, safety_params):
        """
        障害物回避のための速度計算
        """
        max_linear = safety_params['max_linear_velocity']
        max_angular = safety_params['max_angular_velocity']
        
        # 距離に基づく線形速度の調整
        linear_scale = min(1.0, (obstacle_distance - safety_params['emergency_distance']) / 
                          (safety_params['safety_distance'] - safety_params['emergency_distance']))
        linear_vel = current_velocity.linear.x * linear_scale
        
        # 障害物の方向に基づく角速度の計算
        angular_vel = max_angular * np.sign(obstacle_angle) * (1 - linear_scale)
        
        return linear_vel, angular_vel

    @staticmethod
    def create_pose(x: float, y: float, yaw: float) -> Pose:
        """位置と向きからPoseメッセージを作成"""
        pose = Pose()
        pose.position = Point(x=x, y=y, z=0.0)
        q = euler2quat(0, 0, yaw)  # quaternion_from_eulerをeuler2quatに変更
        pose.orientation = Quaternion(x=q[0], y=q[1], z=q[2], w=q[3])
        return pose

    @staticmethod
    def get_yaw_from_pose(pose: Pose) -> float:
        """クォータニオンからヨー角を取得（修正版）"""
        # クォータニオンの要素を取得
        x = pose.orientation.x
        y = pose.orientation.y
        z = pose.orientation.z
        w = pose.orientation.w
    
        # ヨー角の計算（オイラー角のz成分）
        siny_cosp = 2.0 * (w * z + x * y)
        cosy_cosp = 1.0 - 2.0 * (y * y + z * z)
    
        # arctan2を使用して-πからπの範囲のヨー角を取得
        yaw = np.arctan2(siny_cosp, cosy_cosp)
    
        return yaw
