#!/usr/bin/env python3
# === robot_control/robot_control/test_monitor.py ===

import rclpy
from rclpy.node import Node
from nav_msgs.msg import Odometry
from std_msgs.msg import Int32MultiArray
from robot_interfaces.msg import RobotStatus
import csv
import time
import os
from datetime import datetime

class TestMonitor(Node):
    def __init__(self):
        super().__init__('test_monitor')
        
        # パラメータの宣言
        self.declare_parameter('test_timeout', 1800.0)
        self.declare_parameter('navigation_tolerance', 0.2)
        self.declare_parameter('angular_tolerance', 0.1)
        self.declare_parameter('min_battery_level', 20.0)
        
        # パラメータの取得
        self.test_timeout = self.get_parameter('test_timeout').value
        self.navigation_tolerance = self.get_parameter('navigation_tolerance').value
        self.angular_tolerance = self.get_parameter('angular_tolerance').value
        self.min_battery_level = self.get_parameter('min_battery_level').value
        
        # サブスクライバーの設定
        self.odom_sub = self.create_subscription(
            Odometry,
            'odom',
            self.odom_callback,
            10)
            
        self.encoder_sub = self.create_subscription(
            Int32MultiArray,
            'encoder',
            self.encoder_callback,
            10)
            
        self.status_sub = self.create_subscription(
            RobotStatus,
            'robot_status',
            self.status_callback,
            10)
            
        # データログの初期化
        self.setup_data_logging()
        
        # 監視タイマーの設定
        self.create_timer(1.0, self.monitor_callback)
        
        # テスト開始時間の記録
        self.start_time = time.time()
        
        self.get_logger().info('Test monitor started')
        
    def setup_data_logging(self):
        """データログの設定"""
        # ログディレクトリの作成
        log_dir = 'test_logs'
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
            
        # ログファイルの設定
        timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
        self.log_file = os.path.join(log_dir, f'test_log_{timestamp}.csv')
        
        # CSVヘッダーの書き込み
        with open(self.log_file, 'w', newline='') as f:
            writer = csv.writer(f)
            writer.writerow([
                'timestamp',
                'x', 'y', 'theta',
                'linear_velocity', 'angular_velocity',
                'encoder_left', 'encoder_right',
                'battery_level', 'system_temperature',
                'error_status'
            ])
        
    def odom_callback(self, msg):
        """オドメトリデータのコールバック"""
        self.log_data(msg=msg, data_type='odom')
        
    def encoder_callback(self, msg):
        """エンコーダデータのコールバック"""
        self.log_data(msg=msg, data_type='encoder')
        
    def status_callback(self, msg):
        """ロボットステータスのコールバック"""
        self.log_data(msg=msg, data_type='status')
        
        # エラーチェック
        if msg.emergency_stop_activated:
            self.get_logger().error('Emergency stop activated!')
            
        if msg.battery_level < self.min_battery_level:
            self.get_logger().warn(f'Low battery: {msg.battery_level}%')
            
        if msg.motor_error:
            self.get_logger().error('Motor error detected!')
            
    def log_data(self, msg, data_type):
        """データのログ記録"""
        try:
            with open(self.log_file, 'a', newline='') as f:
                writer = csv.writer(f)
                timestamp = time.time()
                
                if data_type == 'odom':
                    writer.writerow([
                        timestamp,
                        msg.pose.pose.position.x,
                        msg.pose.pose.position.y,
                        0.0,  # theta（四元数から計算が必要）
                        msg.twist.twist.linear.x,
                        msg.twist.twist.angular.z,
                        '', '', '', '', ''  # 他のデータ用の空欄
                    ])
                elif data_type == 'encoder':
                    writer.writerow([
                        timestamp,
                        '', '', '', '', '',
                        msg.data[0], msg.data[1],
                        '', '', ''
                    ])
                elif data_type == 'status':
                    writer.writerow([
                        timestamp,
                        '', '', '', '', '', '', '',
                        msg.battery_level,
                        msg.system_temperature,
                        'Error' if msg.is_error else 'OK'
                    ])
        except Exception as e:
            self.get_logger().error(f'Failed to log data: {str(e)}')
            
    def monitor_callback(self):
        """定期的な監視処理"""
        # テストタイムアウトのチェック
        if time.time() - self.start_time > self.test_timeout:
            self.get_logger().error('Test timeout!')
            rclpy.shutdown()
            
def main(args=None):
    rclpy.init(args=args)
    node = TestMonitor()
    try:
        rclpy.spin(node)
    except KeyboardInterrupt:
        pass
    finally:
        node.destroy_node()
        rclpy.shutdown()

if __name__ == '__main__':
    main()