#!/usr/bin/env python3

import copy
import rclpy
from rclpy.node import Node
from nav_msgs.msg import Odometry
from geometry_msgs.msg import Twist, PoseStamped
from std_msgs.msg import Int32MultiArray
from sensor_msgs.msg import LaserScan  # LiDARデータ用
from robot_interfaces.msg import RobotStatus, EmergencyStop, NavigationGoal
from robot_control.robot_utilities import RobotUtilities
from nav2_msgs.action import NavigateToPose
from rclpy.action import ActionClient
import numpy as np

class RobotNavigationControl(Node):
    def __init__(self):
        super().__init__('robot_navigation_control')
        
        self.declare_parameters(
            namespace='',
            parameters=[
                ('max_linear_velocity', 0.5),
                ('max_angular_velocity', 1.0),
                ('position_tolerance', 0.1),
                ('orientation_tolerance', 0.1),
                ('safety_distance', 0.3),        # 障害物との最小安全距離
                ('emergency_distance', 0.2),      # 緊急停止距離
                ('scan_angle_min', -180.0),       # スキャン角度範囲（度）
                ('scan_angle_max', 180.0),
            ]
        )
        
        self.max_linear_velocity = self.get_parameter('max_linear_velocity').value
        self.max_angular_velocity = self.get_parameter('max_angular_velocity').value
        self.position_tolerance = self.get_parameter('position_tolerance').value
        self.orientation_tolerance = self.get_parameter('orientation_tolerance').value
        self.safety_distance = self.get_parameter('safety_distance').value
        self.emergency_distance = self.get_parameter('emergency_distance').value
        self.scan_angle_min = np.radians(self.get_parameter('scan_angle_min').value)
        self.scan_angle_max = np.radians(self.get_parameter('scan_angle_max').value)
        
        self.cmd_vel_pub = self.create_publisher(Twist, 'cmd_vel', 10)
        self.robot_status_pub = self.create_publisher(RobotStatus, 'robot_status', 10)
        
        self.odom_sub = self.create_subscription(
            Odometry, 
            'odom', 
            self.odom_callback, 
            10
        )
        self.encoder_sub = self.create_subscription(
            Int32MultiArray, 
            'encoder', 
            self.encoder_callback, 
            10
        )
        self.emergency_sub = self.create_subscription(
            EmergencyStop, 
            'emergency_stop', 
            self.emergency_callback, 
            10
        )
        self.goal_sub = self.create_subscription(
            NavigationGoal, 
            'navigation_goal', 
            self.goal_callback, 
            10
        )
        self.scan_sub = self.create_subscription(
            LaserScan,
            'scan',
            self.scan_callback,
            10
        )
            
        self.nav_client = ActionClient(self, NavigateToPose, 'navigate_to_pose')
        
        self.current_pose = None
        self.current_velocity = None
        self.emergency_stop = False
        self.navigation_active = False
        self.goal_pose = None
        
        self.obstacle_detected = False
        self.min_distance = float('inf')
        self.scan_data = None
        
        self.encoder_left = 0
        self.encoder_right = 0
        
        self.create_timer(0.1, self.control_loop)
        
        self.get_logger().info('Robot Navigation Control node initialized')
    
    def odom_callback(self, msg):
        self.current_pose = msg.pose.pose
        self.current_velocity = msg.twist.twist
        
    def encoder_callback(self, msg):
        self.encoder_left = msg.data[0]
        self.encoder_right = msg.data[1]
      
    def scan_callback(self, msg):
        """LiDARデータの処理"""
        self.scan_data = self._filter_pole_from_scan(msg)  
        
        # 前方の範囲のデータのみを使用
        start_idx = int((self.scan_angle_min - msg.angle_min) / msg.angle_increment)
        end_idx = int((self.scan_angle_max - msg.angle_min) / msg.angle_increment)
        
        # 有効な範囲内のデータを抽出
        valid_ranges = [r for r in msg.ranges[start_idx:end_idx] 
                       if msg.range_min <= r <= msg.range_max]
        
        if valid_ranges:
            self.min_distance = min(valid_ranges)
            self.obstacle_detected = self.min_distance < self.safety_distance
            
            # 緊急停止距離以下の場合
            if self.min_distance < self.emergency_distance:
                self.emergency_stop = True
                self.stop_robot()
                self.get_logger().warn(f'Emergency stop: Obstacle detected at {self.min_distance:.2f}m')

    def _filter_pole_from_scan(self, scan_msg):
        """スキャンデータからポールを除外"""
        filtered_scan = copy.deepcopy(scan_msg)
        pole_distance = 0.35  # ポールまでの距離
        pole_tolerance = 0.05  # 許容誤差
        
        # 各スキャンデータについてポール位置のデータを除外
        for i, r in enumerate(filtered_scan.ranges):
            if abs(r - pole_distance) < pole_tolerance:
                # ポール位置のデータを無効化（前後の値で補間）
                prev_idx = (i - 1) % len(filtered_scan.ranges)
                next_idx = (i + 1) % len(filtered_scan.ranges)
                if filtered_scan.ranges[prev_idx] != float('inf') and filtered_scan.ranges[next_idx] != float('inf'):
                    filtered_scan.ranges[i] = (filtered_scan.ranges[prev_idx] + filtered_scan.ranges[next_idx]) / 2
                else:
                    filtered_scan.ranges[i] = float('inf')
        
        return filtered_scan
  
    def emergency_callback(self, msg):
        self.emergency_stop = msg.stop
        if self.emergency_stop:
            self.stop_robot()
            self.get_logger().warn('Emergency stop activated!')
        
    def goal_callback(self, msg):
        if self.emergency_stop:
            self.get_logger().warn('Cannot start navigation while emergency stop is active')
            return
            
        self.goal_pose = msg
        self.start_navigation()
        
    def start_navigation(self):
        if not self.nav_client.wait_for_server(timeout_sec=1.0):
            self.get_logger().error('Navigation action server not available')
            return
            
        goal_msg = NavigateToPose.Goal()
        goal_msg.pose.header.frame_id = 'map'
        goal_msg.pose.header.stamp = self.get_clock().now().to_msg()
        goal_msg.pose.pose = self.goal_pose.pose
        
        self.navigation_active = True
        self.nav_client.send_goal_async(goal_msg, feedback_callback=self.navigation_feedback_callback)
            
    def navigation_feedback_callback(self, feedback_msg):
        feedback = feedback_msg.feedback
        status_msg = RobotStatus()
        status_msg.navigation_active = self.navigation_active
        status_msg.current_pose = self.current_pose
        status_msg.distance_remaining = feedback.distance_remaining
        self.robot_status_pub.publish(status_msg)
        
    def stop_robot(self):
        stop_msg = Twist()
        stop_msg.linear.x = 0.0
        stop_msg.angular.z = 0.0
        self.cmd_vel_pub.publish(stop_msg)
        self.navigation_active = False
        
    def velocity_controller(self):
        """速度制御と衝突回避の統合"""
        if not self.scan_data or not self.current_pose:
            return Twist()
            
        safety_params = {
            'emergency_distance': self.emergency_distance,
            'safety_distance': self.safety_distance,
            'max_linear_velocity': self.max_linear_velocity,
            'max_angular_velocity': self.max_angular_velocity
        }
        
        # 安全性チェック
        is_safe, min_dist, obstacle_angle = RobotUtilities.check_safety_zone(
            self.scan_data, safety_params)
        
        cmd_vel = Twist()
        if not is_safe:
            # 障害物回避の速度計算
            linear_vel, angular_vel = RobotUtilities.calculate_avoidance_velocity(
                self.current_velocity, min_dist, obstacle_angle, safety_params)
            
            cmd_vel.linear.x = linear_vel
            cmd_vel.angular.z = angular_vel
            self.get_logger().info(f'Obstacle avoidance: linear={linear_vel:.2f}, angular={angular_vel:.2f}')
        else:
            # 通常のナビゲーション速度
            cmd_vel.linear.x = self.max_linear_velocity
            cmd_vel.angular.z = 0.0
        
        return cmd_vel

    def control_loop(self):
        if self.emergency_stop:
            self.stop_robot()
            return
                
        if not self.navigation_active:
            return
                
        # 速度制御の実行
        cmd_vel = self.velocity_controller()
        self.cmd_vel_pub.publish(cmd_vel)
        
        # ステータス更新
        status_msg = RobotStatus()
        status_msg.navigation_active = self.navigation_active
        status_msg.current_pose = self.current_pose
        status_msg.emergency_stop = self.emergency_stop
        status_msg.obstacle_detected = self.obstacle_detected
        status_msg.min_distance = self.min_distance if self.min_distance != float('inf') else 0.0
        self.robot_status_pub.publish(status_msg)

def main(args=None):
    rclpy.init(args=args)
    node = RobotNavigationControl()
    try:
        rclpy.spin(node)
    except KeyboardInterrupt:
        pass
    finally:
        node.destroy_node()
        rclpy.shutdown()

if __name__ == '__main__':
    main()
