#!/usr/bin/env python3
# === robot_control/scripts/analyze_test_results.py ===

import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import argparse
import os
import yaml
from datetime import datetime

def load_test_data(log_file):
    """テストログデータの読み込み"""
    try:
        df = pd.read_csv(log_file)
        df['timestamp'] = pd.to_datetime(df['timestamp'], unit='s')
        return df
    except Exception as e:
        print(f"Error loading test data: {str(e)}")
        return None

def analyze_navigation_performance(df):
    """ナビゲーション性能の分析"""
    results = {
        'mean_linear_velocity': df['linear_velocity'].mean(),
        'max_linear_velocity': df['linear_velocity'].max(),
        'mean_angular_velocity': df['angular_velocity'].mean(),
        'max_angular_velocity': abs(df['angular_velocity']).max()
    }
    
    # 位置追従誤差の計算
    if 'target_x' in df.columns and 'target_y' in df.columns:
        df['position_error'] = ((df['x'] - df['target_x'])**2 + 
                              (df['y'] - df['target_y'])**2)**0.5
        results['mean_position_error'] = df['position_error'].mean()
        results['max_position_error'] = df['position_error'].max()
    
    return results

def plot_results(df, output_dir):
    """結果のプロット作成"""
    # 出力ディレクトリの作成
    os.makedirs(output_dir, exist_ok=True)
    
    # 軌跡プロット
    plt.figure(figsize=(10, 10))
    plt.plot(df['x'], df['y'], 'b-', label='Robot Path')
    plt.grid(True)
    plt.xlabel('X [m]')
    plt.ylabel('Y [m]')
    plt.title('Robot Navigation Path')
    plt.legend()
    plt.savefig(os.path.join(output_dir, 'navigation_path.png'))
    
    # 速度プロット
    plt.figure(figsize=(12, 6))
    plt.plot(df['timestamp'], df['linear_velocity'], 'g-', label='Linear Velocity')
    plt.plot(df['timestamp'], df['angular_velocity'], 'r-', label='Angular Velocity')
    plt.grid(True)
    plt.xlabel('Time')
    plt.ylabel('Velocity')
    plt.title('Robot Velocities')
    plt.legend()
    plt.savefig(os.path.join(output_dir, 'velocities.png'))
    
    # システム状態プロット
    if 'battery_level' in df.columns and 'system_temperature' in df.columns:
        plt.figure(figsize=(12, 6))
        plt.plot(df['timestamp'], df['battery_level'], 'b-', label='Battery Level')
        plt.plot(df['timestamp'], df['system_temperature'], 'r-', label='Temperature')
        plt.grid(True)
        plt.xlabel('Time')
        plt.ylabel('Value')
        plt.title('System Status')
        plt.legend()
        plt.savefig(os.path.join(output_dir, 'system_status.png'))

def generate_report(results, output_dir):
    """テストレポートの生成"""
    report_path = os.path.join(output_dir, 'test_report.html')
    
    html_content = f"""
    <html>
    <head>
        <title>Navigation Test Report</title>
        <style>
            body {{ font-family: Arial, sans-serif; margin: 20px; }}
            table {{ border-collapse: collapse; width: 100%; }}
            th, td {{ border: 1px solid #ddd; padding: 8px; text-align: left; }}
            th {{ background-color: #f2f2f2; }}
            img {{ max-width: 800px; }}
        </style>
    </head>
    <body>
        <h1>Navigation Test Report</h1>
        <h2>Test Results Summary</h2>
        <table>
            <tr><th>Metric</th><th>Value</th></tr>
    """
    
    for key, value in results.items():
        html_content += f"<tr><td>{key}</td><td>{value:.3f}</td></tr>"
    
    html_content += """
        </table>
        <h2>Navigation Path</h2>
        <img src="navigation_path.png" alt="Navigation Path">
        <h2>Velocity Profile</h2>
        <img src="velocities.png" alt="Velocity Profile">
        <h2>System Status</h2>
        <img src="system_status.png" alt="System Status">
    </body>
    </html>
    """
    
    with open(report_path, 'w') as f:
        f.write(html_content)

def main():
    parser = argparse.ArgumentParser(description='Analyze navigation test results')
    parser.add_argument('log_file', help='Path to test log file')
    parser.add_argument('--output', help='Output directory for results',
                      default='test_results')
    args = parser.parse_args()
    
    # データの読み込みと分析
    df = load_test_data(args.log_file)
    if df is None:
        return
        
    results = analyze_navigation_performance(df)
    
    # 結果の出力
    timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
    output_dir = os.path.join(args.output, f'analysis_{timestamp}')
    os.makedirs(output_dir, exist_ok=True)
    
    plot_results(df, output_dir)
    generate_report(results, output_dir)

def analyze_system_stability(df):
    """システムの安定性分析"""
    stability_metrics = {
        'velocity_stability': {
            'linear_std': df['linear_velocity'].std(),
            'angular_std': df['angular_velocity'].std(),
            'sudden_changes': len(df[abs(df['linear_velocity'].diff()) > 0.5])
        },
        'encoder_reliability': {
            'left_jumps': len(df[abs(df['encoder_left'].diff()) > 100]),
            'right_jumps': len(df[abs(df['encoder_right'].diff()) > 100])
        },
        'system_health': {
            'min_battery': df['battery_level'].min(),
            'max_temperature': df['system_temperature'].max(),
            'error_count': len(df[df['error_status'] == 'Error'])
        }
    }
    return stability_metrics

def export_test_summary(results, stability_metrics, output_dir):
    """テスト結果のサマリーをエクスポート"""
    summary_path = os.path.join(output_dir, 'test_summary.yaml')
    
    summary = {
        'test_date': datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
        'navigation_performance': results,
        'system_stability': stability_metrics,
        'overall_status': 'PASS' if _evaluate_test_success(results, stability_metrics) else 'FAIL'
    }
    
    with open(summary_path, 'w') as f:
        yaml.dump(summary, f, default_flow_style=False)
        
def _evaluate_test_success(results, stability_metrics):
    """テスト成功の評価"""
    # 成功条件の定義
    criteria = {
        'max_position_error': 0.5,  # 最大位置誤差 0.5m
        'min_battery_level': 20.0,  # 最小バッテリー残量 20%
        'max_temperature': 70.0,    # 最大温度 70℃
        'max_error_count': 5        # 最大エラー数
    }
    
    # 条件チェック
    checks = [
        results.get('max_position_error', 0) < criteria['max_position_error'],
        stability_metrics['system_health']['min_battery'] > criteria['min_battery_level'],
        stability_metrics['system_health']['max_temperature'] < criteria['max_temperature'],
        stability_metrics['system_health']['error_count'] < criteria['max_error_count']
    ]
    
    return all(checks)

if __name__ == '__main__':
    main()