#!/usr/bin/env python3
# === robot_control/scripts/setup_test.py ===

import argparse
import os
import yaml
import subprocess
import time

def check_robot_connection():
    """ロボットとの接続確認"""
    try:
        result = subprocess.run(['ros2', 'topic', 'list'], 
                              capture_output=True, text=True)
        required_topics = ['/odom', '/cmd_vel', '/encoder', '/robot_status']
        for topic in required_topics:
            if topic not in result.stdout:
                print(f"Warning: Required topic {topic} not found")
                return False
        return True
    except subprocess.CalledProcessError:
        print("Error: Failed to check ROS2 topics")
        return False

def verify_map_file(map_path):
    """マップファイルの検証"""
    if not os.path.exists(map_path):
        print(f"Error: Map file not found at {map_path}")
        return False
        
    try:
        with open(map_path, 'r') as f:
            map_data = yaml.safe_load(f)
            required_fields = ['image', 'resolution', 'origin', 'occupied_thresh', 'free_thresh']
            for field in required_fields:
                if field not in map_data:
                    print(f"Error: Required field '{field}' missing in map file")
                    return False
        return True
    except Exception as e:
        print(f"Error reading map file: {str(e)}")
        return False

def setup_test_environment():
    """テスト環境のセットアップ"""
    # テストディレクトリの作成
    os.makedirs('test_logs', exist_ok=True)
    os.makedirs('test_results', exist_ok=True)
    
    # パーミッションの設定
    try:
        subprocess.run(['sudo', 'chmod', '666', '/dev/ttyUSB0'])
    except subprocess.CalledProcessError:
        print("Warning: Failed to set USB port permissions")

def main():
    parser = argparse.ArgumentParser(description='Setup navigation test environment')
    parser.add_argument('--map', type=str, help='Path to map file',
                      default='maps/school_map.yaml')
    parser.add_argument('--check-only', action='store_true',
                      help='Only check the setup without making changes')
    args = parser.parse_args()
    
    print("Setting up test environment...")
    
    # 環境チェック
    if not check_robot_connection():
        print("Error: Robot connection check failed")
        return False
        
    if not verify_map_file(args.map):
        print("Error: Map file verification failed")
        return False
        
    if args.check_only:
        print("Check completed successfully")
        return True
        
    # テスト環境のセットアップ
    setup_test_environment()
    
    print("Test environment setup completed successfully")
    return True

if __name__ == '__main__':
    main()