#!/bin/bash
# === robot_control/scripts/setup_test.sh ===

# エラーハンドリング
set -e

# 色の定義
RED='\033[0;31m'
GREEN='\033[0;32m'
YELLOW='\033[1;33m'
NC='\033[0m'

echo -e "${YELLOW}Starting navigation test setup...${NC}"

# ROSワークスペースの確認
if [ -z "$ROS_DISTRO" ]; then
    echo -e "${RED}Error: ROS2 environment not sourced${NC}"
    exit 1
fi

# 依存パッケージの確認
echo "Checking required packages..."
required_packages=(
    "nav2_bringup"
    "micro_ros_agent"
    "robot_control"
    "my_school_map"
)

for package in "${required_packages[@]}"; do
    if ! ros2 pkg list | grep -q "^$package\$"; then
        echo -e "${RED}Error: Package $package not found${NC}"
        exit 1
    fi
done

# シリアルポートの設定
echo "Setting up serial port..."
if [ -e "/dev/ttyUSB0" ]; then
    sudo chmod 666 /dev/ttyUSB0
else
    echo -e "${RED}Error: Serial port /dev/ttyUSB0 not found${NC}"
    exit 1
fi

# テストディレクトリの準備
echo "Preparing test directories..."
mkdir -p test_logs test_results

# ログファイルのクリーンアップ
echo "Cleaning up old log files..."
find test_logs -type f -mtime +7 -exec rm {} \;

# 設定ファイルの検証
echo "Validating configuration files..."
config_files=(
    "config/navigation_params.yaml"
    "maps/school_map.yaml"
)

for config in "${config_files[@]}"; do
    if [ ! -f "$config" ]; then
        echo -e "${RED}Error: Configuration file $config not found${NC}"
        exit 1
    fi
done

# ROSノードの状態確認
echo "Checking ROS nodes..."
if ! ros2 node list > /dev/null 2>&1; then
    echo -e "${YELLOW}Warning: No ROS nodes running${NC}"
fi

# テスト実行権限の設定
echo "Setting up test permissions..."
chmod +x scripts/analyze_test_results.py
chmod +x scripts/setup_test.py

echo -e "${GREEN}Test setup completed successfully${NC}"

# 使用方法の表示
echo -e "\nUsage:"
echo "1. Start the test environment:"
echo "   ros2 launch robot_control test_navigation.launch.py"
echo "2. Run the tests:"
echo "   ros2 test robot_control"
echo "3. Analyze results:"
echo "   ./scripts/analyze_test_results.py test_logs/latest.csv"