import socket
import json
import threading
import time
from datetime import datetime
from geometry_msgs.msg import Pose
import math

class NavigationClient:
    def __init__(self, host='localhost', port=5001):
        self.host = host
        self.port = port
        self.sock = None
        self.is_running = True
        self.current_pose = None
        self.new_destination = None
        self.connection_event = threading.Event()
        
        # 再接続パラメータ
        self.max_retries = 10
        self.retry_delay = 1
        self.connection_timeout = 30
        self.receive_timeout = 0.1
        
        # 統計情報
        self.connection_attempts = 0
        self.last_successful_send = None
        self.successful_sends = 0
        self.failed_sends = 0
        
        # デバッグモード
        self.debug_mode = True
        
        # 通信スレッド開始
        self.thread = threading.Thread(target=self.run)
        self.thread.daemon = True
        self.thread.start()
        
    def debug_print(self, message, data=None):
        """デバッグ情報の出力"""
        if self.debug_mode:
            timestamp = datetime.now().strftime('%H:%M:%S.%f')[:-3]
            debug_msg = f"[{timestamp}] {message}"
            if data:
                debug_msg += f": {data}"
            print(debug_msg)
        
    def connect_with_retry(self):
        """再試行付きの接続処理"""
        retry_count = 0
        while self.is_running and retry_count < self.max_retries:
            try:
                if self.sock:
                    try:
                        self.sock.close()
                    except:
                        pass
                
                self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                
                # ソケットオプションの設定
                self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 8192)
                self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 8192)
                self.sock.settimeout(10)  # 接続タイムアウト
                
                self.debug_print(f"サーバー({self.host}:{self.port})に接続を試みます... 試行回数: {retry_count + 1}")
                self.sock.connect((self.host, self.port))
                self.debug_print("接続成功")
                
                # 接続成功後、受信タイムアウトを設定
                self.sock.settimeout(self.receive_timeout)
                self.connection_event.set()
                self.connection_attempts += 1
                return True
                
            except Exception as e:
                self.debug_print(f"接続エラー", str(e))
                retry_count += 1
                if retry_count < self.max_retries:
                    wait_time = self.retry_delay * retry_count
                    self.debug_print(f"{wait_time}秒後に再試行します")
                    time.sleep(wait_time)
                else:
                    self.debug_print("最大再試行回数を超えました")
                    return False
        return False
        
    def send_position(self, pose: Pose):
        """現在位置の送信"""
        if not self.sock or not self.connection_event.is_set():
            return
            
        try:
            # 送信タイムアウトの設定
            self.sock.settimeout(5.0)
            
            position_data = {
                "type": "position",
                "data": {
                    "x": pose.position.x,
                    "y": pose.position.y
                },
                "timestamp": time.time()
            }
            
            json_data = json.dumps(position_data)
            self.sock.send(json_data.encode('utf-8'))
            
            # 送信成功の記録
            self.last_successful_send = time.time()
            self.successful_sends += 1
            self.debug_print("位置情報送信成功", position_data["data"])
            
            # 送信後にタイムアウトを元に戻す
            self.sock.settimeout(self.receive_timeout)
            
        except socket.timeout:
            self.debug_print("送信タイムアウト")
            self.failed_sends += 1
            self.connection_event.clear()
        except Exception as e:
            self.debug_print("送信エラー", str(e))
            self.failed_sends += 1
            self.connection_event.clear()
            
    def handle_receive(self):
        """データ受信処理"""
        if not self.sock:
            return
        
        try:
            data = self.sock.recv(1024)
            if data:
                try:
                    # バッファ全体をデコード
                    decoded_data = data.decode('utf-8').strip()
                    self.debug_print("受信データ", decoded_data)
                
                    # 複数のJSONメッセージを処理
                    messages = decoded_data.split("}{")
                    for i, msg in enumerate(messages):
                        # メッセージの修復
                        if i > 0:
                            msg = "{" + msg
                        if i < len(messages) - 1:
                            msg = msg + "}"
                    
                        try:
                            json_data = json.loads(msg)
                        
                            if json_data["type"] == "destination":
                                dest_data = json_data["data"]
                                new_pose = Pose()
                                new_pose.position.x = float(dest_data["x"])
                                new_pose.position.y = float(dest_data["y"])
                                self.new_destination = new_pose
                                self.debug_print("新しい目的地を設定", 
                                    {"x": new_pose.position.x, "y": new_pose.position.y})
                                
                        except json.JSONDecodeError as e:
                            self.debug_print("個別のJSON解析エラー", str(e))
                            continue
                        
                except json.JSONDecodeError as e:
                    self.debug_print("JSON解析エラー", str(e))
                except Exception as e:
                    self.debug_print("メッセージ処理エラー", str(e))
                    
        except socket.timeout:
            pass
        except Exception as e:
            self.debug_print("受信エラー", str(e))
            self.connection_event.clear()
            
    def run(self):
        """メインループ"""
        if not self.connect_with_retry():
            self.debug_print("初期接続に失敗しました")
            return
            
        while self.is_running:
            try:
                if not self.connection_event.is_set():
                    self.debug_print("接続が切断されています。再接続を試みます...")
                    if not self.connect_with_retry():
                        time.sleep(self.retry_delay)
                        continue
                        
                # 現在位置の送信
                if self.current_pose:
                    self.send_position(self.current_pose)
                    self.debug_print("現在位置を送信", 
                        {"x": self.current_pose.position.x, 
                         "y": self.current_pose.position.y})
                    
                # データ受信
                self.handle_receive()
                
                time.sleep(0.1)
                
            except Exception as e:
                self.debug_print("予期せぬエラー", str(e))
                self.connection_event.clear()
                time.sleep(0.5)
                
    def update_position(self, pose: Pose):
        """位置情報の更新"""
        self.current_pose = pose
        
    def get_new_destination(self) -> Pose:
        """新しい目的地の取得"""
        if self.new_destination:
            dest = self.new_destination
            self.new_destination = None
            return dest
        return None
        
    def cleanup(self):
        """終了処理"""
        self.debug_print("\n終了処理を開始します...")
    
        # 実行フラグを先に落とす
        self.is_running = False
        time.sleep(0.1)  # スレッドの終了を待つ
    
        if self.sock:
            try:
                # 送信バッファをフラッシュ
                self.sock.shutdown(socket.SHUT_WR)
                time.sleep(0.1)  # 送信完了を待つ
            
                # 受信バッファをクリア
                self.sock.settimeout(0.5)
                try:
                    while self.sock.recv(1024):
                        pass
                except socket.timeout:
                    pass
                except Exception:
                    pass
                
                # ソケットを閉じる
                self.sock.close()
            
            except Exception as e:
                self.debug_print(f"ソケットのクリーンアップ中にエラー: {str(e)}")
    
        try:
            print("\n統計情報:")
            print(f"- 接続試行回数: {self.connection_attempts}")
            print(f"- 成功した送信: {self.successful_sends}")
            print(f"- 失敗した送信: {self.failed_sends}")
            if self.last_successful_send:
                print(f"- 最後の成功送信: {datetime.fromtimestamp(self.last_successful_send).strftime('%Y-%m-%d %H:%M:%S')}")
        except Exception as e:
            self.debug_print(f"統計情報出力中にエラー: {str(e)}")
        
    print("通信を終了します")
