import unittest
import rclpy
from geometry_msgs.msg import Pose, Point, PoseStamped, Twist
from nav_msgs.msg import Path, Odometry
from sensor_msgs.msg import LaserScan
from std_msgs.msg import Int32MultiArray
from robot_control.path_planner import PathPlanner
from robot_control.robot_utilities import RobotUtilities
from robot_control.robot_navigation_control import RobotNavigationControl
import math
import yaml
import os
import numpy as np
import time
from rclpy.callback_groups import ReentrantCallbackGroup
from rclpy.executors import MultiThreadedExecutor
from rclpy.qos import QoSProfile, ReliabilityPolicy, HistoryPolicy

from navigation_client import NavigationClient

class TestLongDistanceNavigation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        rclpy.init()
        
        # read map data
        map_path = os.path.join(os.path.dirname(__file__), '../../my_school_map/maps/school_map.yaml')
        with open(map_path, 'r') as f:
            cls.map_config = yaml.safe_load(f)
            
        cls.executor = MultiThreadedExecutor()
        
    def setUp(self):
        self.goal_pose = None
        self.navigation_client = NavigationClient(host='172.25.17.32', port=5001)
        self.setup_test_environment()
        
    def setup_test_environment(self):
        """setup test environment"""
        self.node = rclpy.create_node('long_distance_navigation_test_node')
        callback_group = ReentrantCallbackGroup()
        self.node.callback_group = callback_group

        # init params
        self.safety_distance = 0.8
        self.emergency_distance = 0.5
        self.min_valid_range = 0.45
        self.max_valid_range = 8.0

        self.qos_profile = QoSProfile(
            reliability=ReliabilityPolicy.RELIABLE,
            history=HistoryPolicy.KEEP_LAST,
            depth=10
        )
        
        # pablisher and subscriver setting
        self.cmd_vel_pub = self.node.create_publisher(
            Twist, 'cmd_vel', 10)
            
        self.odom_sub = self.node.create_subscription(
            Odometry,
            'odom',
            self.odom_callback,
            10)
            
        self.encoder_sub = self.node.create_subscription(
            Int32MultiArray,
            'encoder',
            self.encoder_callback,
            10)
            
        self.scan_sub = self.node.create_subscription(
            LaserScan,
            'scan',
            self.scan_callback,
            10)
        
        # navigation conpornent setting
        self.planner = PathPlanner()
        self.planner.resolution = 0.1
        self.planner.robot_radius = 0.4
        self.utils = RobotUtilities()
        self.controller = RobotNavigationControl()
        
        # stable variable setting
        self.current_odom = None
        self.current_scan = None
        self.encoder_counts = None
        self.obstacle_detected = False
        self.min_distance = float('inf')
        
        # calibration wait time
        self._wait_for_system_ready()

    def _wait_for_system_ready(self):
        """wait for sistem setup"""
        timeout = time.time() + 30.0  # 30sec timeout
        
        while not (self.current_odom and self.encoder_counts):
            if time.time() > timeout:
                raise TimeoutError("did not setup sistems")
            rclpy.spin_once(self.node, timeout_sec=0.1)

    def odom_callback(self, msg):
        self.current_odom = msg
        self.navigation_client.update_position(msg.pose.pose)

    def encoder_callback(self, msg):
        self.encoder_counts = msg.data

    def scan_callback(self, msg):
        self.current_scan = msg
        
        # front data filtaring
        start_idx = int((np.radians(45) - msg.angle_min) / msg.angle_increment)
        end_idx = int((np.radians(135) - msg.angle_min) / msg.angle_increment)
        
        # index lange check
        start_idx = max(0, start_idx)
        end_idx = min(len(msg.ranges), end_idx)
        
        # length data setting
        valid_ranges = []
        for i in range(start_idx, end_idx):
            r = msg.ranges[i]
            if (np.isfinite(r) and 
                self.min_valid_range <= r <= self.max_valid_range):
                valid_ranges.append(r)
        
        if valid_ranges:
            self.min_distance = min(valid_ranges)
            self.obstacle_detected = self.min_distance < self.safety_distance

    def test_remote_navigation(self):
        try:
            if not self._verify_initial_position():
                self.skipTest("robot position is not correct to initioal position")
            
            while True:
                self.node.get_logger().info("waiting for setting goals")
                
                new_goal = self._wait_for_new_goal()
                if not new_goal:
                    continue
                    
                # goal position setting
                self.goal_pose = new_goal
                
                self.node.get_logger().info(
                    f"get new goals: x={new_goal.position.x}, y={new_goal.position.y}")
                
                success = self._execute_navigation_to_goal(new_goal)
                
                if success:
                    self._verify_navigation_data(new_goal)
                    self.node.get_logger().info("arriving at goal")
                else:
                    self.node.get_logger().error("false to navigation")
                
                time.sleep(1.0)
                
        except KeyboardInterrupt:
            self.node.get_logger().info("end navigation")

    def _wait_for_new_goal(self, timeout_sec=60.0):
        timeout = time.time() + timeout_sec
        while time.time() < timeout:
            new_goal = self.navigation_client.get_new_destination()
            if new_goal:
                return new_goal
            rclpy.spin_once(self.node, timeout_sec=0.1)
            time.sleep(0.1)
        return None

    def _execute_navigation_to_goal(self, goal_pose: Pose) -> bool:
        try:
            start_time = time.time()
            last_cmd_time = time.time()
            start_encoder_count = list(self.encoder_counts) if self.encoder_counts else None
            timeout = 1200.0  # 10min timeout
            cmd_interval = 0.04  # 20ms command (50Hz)
            
            self.node.get_logger().info(
                f'start running: x={goal_pose.position.x}, y={goal_pose.position.y}')
            
            while not self._reached_goal(goal_pose):
                current_time = time.time()
                
                if current_time - start_time > timeout:
                    self.node.get_logger().error("Navigation timeout")
                    return False

                # cmd_vel update
                if current_time - last_cmd_time >= cmd_interval:
                    current_path = self.planner.create_path(
                        self.current_odom.pose.pose,
                        goal_pose,
                        []
                    )
                    
                    if not current_path:
                        self.node.get_logger().error("Path planning failed")
                        return False
                    
                    cmd_vel = self._calculate_velocity_command(current_path)
                    self.cmd_vel_pub.publish(cmd_vel)
                    last_cmd_time = current_time
                    
                    # encoder get
                    if not self._verify_encoder_movement(start_encoder_count):
                        self.node.get_logger().warn("Encoder movement not detected")
                
                # spin high speed
                rclpy.spin_once(self.node, timeout_sec=0.01)
                
            return True
            
        except Exception as e:
            self.node.get_logger().error(f"Navigation error: {str(e)}")
            return False

    def _calculate_velocity_command(self, path: Path) -> Twist:
        cmd_vel = Twist()
        
        # emangency stop
        if self.min_distance <= self.emergency_distance:
            self.node.get_logger().warn(f'Emergency stop: Obstacle detected at {self.min_distance:.2f}m')
            return cmd_vel  # setting velocity zero

        if not path.poses:
            return cmd_vel

        current_pose = self.current_odom.pose.pose
    
        # caluculate position
        dx = round(self.goal_pose.position.x - current_pose.position.x, 6)
        dy = round(self.goal_pose.position.y - current_pose.position.y, 6)
        distance = np.sqrt(dx*dx + dy*dy)

        # caluculate yaw（-π< yaw <π）
        current_yaw = self.utils.get_yaw_from_pose(current_pose)
        
        if (self.goal_pose.position.y - current_pose.position.y) > 0:
            # goal angle（north = -π/2）
            target_yaw = np.pi/2
        elif (self.goal_pose.position.y - current_pose.position.y) < 0:
            # goal angle（south = -π/2）
            target_yaw = -np.pi/2

        # caluculate target yaw（-π< yaw <π）
        yaw_error = target_yaw - current_yaw
        while yaw_error > np.pi:
            yaw_error -= 2 * np.pi
        while yaw_error < -np.pi:
            yaw_error += 2 * np.pi

        self.node.get_logger().info(
            f'\nNavigation Debug:'
            f'\nCurrent position: x={current_pose.position.x:.3f}, y={current_pose.position.y:.3f}'
            f'\nGoal position: x={self.goal_pose.position.x:.3f}, y={self.goal_pose.position.y:.3f}'
            f'\nDelta: dx={dx:.3f}m, dy={dy:.3f}m'
            f'\nDistance to goal: {distance:.3f}m'
            f'\nCurrent Yaw: {current_yaw:.4f} rad ({np.degrees(current_yaw):.1f} deg)'
            f'\nTarget Yaw: {target_yaw:.4f} rad ({np.degrees(target_yaw):.1f} deg)'
            f'\nYaw Error: {yaw_error:.4f} rad ({np.degrees(yaw_error):.1f} deg)'
        )

        # error angle
        YAW_TOLERANCE = np.radians(10)  # 10 degree
        X_TOLERANCE = 0.15  # 15cm
        
        # angle setting when find obstacle
        speed_scale = 1.0
        if self.obstacle_detected:
            # verocity change from safety_distance to emergency_distance
            speed_scale = max(0.0, min(1.0, 
                (self.min_distance - self.emergency_distance) / 
                (self.safety_distance - self.emergency_distance)))


        if abs(yaw_error) > YAW_TOLERANCE:
            # rotation mode
            base_speed = 0.3  # nomal rotation speed
            scale_factor = min(1.0, abs(yaw_error) / (np.pi/4))
            cmd_vel.angular.z = base_speed * scale_factor * np.sign(yaw_error)
            cmd_vel.linear.x = 0.0
        
            self.node.get_logger().info(
                f'ROTATION MODE:'
                f'\n  yaw_error: {yaw_error:.3f}'
                f'\n  scale_factor: {scale_factor:.3f}'
                f'\n  angular.z: {cmd_vel.angular.z:.3f}'
            )
        else:
            # straight mode
            # speed control
            linear_speed = min(0.3, max(0.1, 0.2 * abs(dy)))
            cmd_vel.linear.x = linear_speed
        
            angular_correction = -np.clip(yaw_error * 0.1, -0.1, 0.1)

            position_correction = -np.clip(dx / X_TOLERANCE * 0.1, -0.1, 0.1)
            cmd_vel.angular.z = angular_correction + position_correction
        
            self.node.get_logger().info(
                f'STRAIGHT MODE:'
                f'\n  linear.x: {cmd_vel.linear.x:.3f}'
                f'\n  angular_correction: {angular_correction:.3f}'
                f'\n  position_correction: {position_correction:.3f}'
                f'\n  total angular.z: {cmd_vel.angular.z:.3f}'
            )

        return cmd_vel
    
    def _reached_goal(self, goal_pose: Pose) -> bool:
        if not self.current_odom:
            return False
        
        current_pose = self.current_odom.pose.pose
        distance = self.utils.calculate_distance(current_pose, goal_pose)
        angle_diff = self.utils.calculate_angle_difference(current_pose, goal_pose)
        
        return distance < 0.5 and abs(angle_diff) < 0.3

    def _verify_initial_position(self) -> bool:
        if not self.current_odom:
            return False
        
        initial_pose = self.current_odom.pose.pose
        return (abs(initial_pose.position.x - 50.5) < 0.5 and 
                abs(initial_pose.position.y - 5.5) < 0.5)

    def _verify_encoder_movement(self, start_counts) -> bool:
        if not start_counts or not self.encoder_counts:
            return True
        
        delta_left = abs(self.encoder_counts[0] - start_counts[0])
        delta_right = abs(self.encoder_counts[1] - start_counts[1])
        
        if delta_left <= 1 and delta_right <= 1:
            self.node.get_logger().warn("Encoder movement not detected")
            return False
        return True

    def _verify_navigation_data(self, goal_pose: Pose):
        self.assertIsNotNone(self.current_odom, "no odometry data")
        self.assertIsNotNone(self.encoder_counts, "no encoder data")
        
        current_pose = self.current_odom.pose.pose
        distance = self.utils.calculate_distance(current_pose, goal_pose)
        angle_diff = self.utils.calculate_angle_difference(current_pose, goal_pose)
        
        self.assertLess(distance, 0.5, "do not reach goal")
        self.assertLess(abs(angle_diff), 0.3, "do not reach target angle")

    def tearDown(self):
        try:
            if hasattr(self, 'cmd_vel_pub'):
                stop_cmd = Twist()
                self.cmd_vel_pub.publish(stop_cmd)
            
            if hasattr(self, 'navigation_client'):
                self.navigation_client.cleanup()
            
            if hasattr(self, 'node'):
                self.node.destroy_node()
            
        except Exception as e:
            print(f"error while cleanup: {str(e)}")

if __name__ == '__main__':
    unittest.main()
