# raspi/audio/audio_client.py
import socketio
import logging
from .audio_handler import RaspberryPiAudio
import time
import threading

logger = logging.getLogger(__name__)

class AudioClient:
    def __init__(self, server_url):
        self.server_url = server_url
        self.sio = socketio.Client(reconnection=True, reconnection_attempts=5)
        self.audio_handler = RaspberryPiAudio()
        self.setup_socket_events()
        self.connected = False
        self.reconnect_attempts = 0
        self.max_reconnect_attempts = 5

    def setup_socket_events(self):
        @self.sio.event
        def connect():
            logger.info("Connected to server")
            self.connected = True
            self.reconnect_attempts = 0
            self.send_status_update()

        @self.sio.event
        def disconnect():
            logger.info("Disconnected from server")
            self.connected = False

        @self.sio.on('raspi_mic_toggle')
        def handle_raspi_mic_toggle(data):
            self.audio_handler.set_mic_enabled(data['enabled'])
            self.send_status_update()

        @self.sio.on('raspi_speaker_toggle')
        def handle_raspi_speaker_toggle(data):
            self.audio_handler.set_speaker_enabled(data['enabled'])
            self.send_status_update()

        @self.sio.on('raspi_volume_change')
        def handle_raspi_volume_change(data):
            self.audio_handler.set_volume(data['volume'])
            self.send_status_update()

        @self.sio.on('speak_text')
        def handle_speak_text(data):
            text = data.get('text', '')
            if text:
                self.audio_handler.text_queue.put(text)
                self.send_status_update(status='speaking')

        @self.sio.on('audio_data')
        def handle_audio_data(data):
            try:
                audio_data = bytes.fromhex(data['audio'])
                self.audio_handler.process_incoming_audio(audio_data)
            except Exception as e:
                logger.error(f"Error processing audio data: {e}")

    def send_status_update(self, status=None):
        """状態更新の送信"""
        try:
            update_data = {
                'mic_enabled': self.audio_handler.state.mic_enabled,
                'speaker_enabled': self.audio_handler.state.speaker_enabled,
                'volume': self.audio_handler.state.volume
            }
            if status:
                update_data['status'] = status
            
            if self.connected:
                self.sio.emit('raspi_status_update', update_data)
        except Exception as e:
            logger.error(f"Failed to send status update: {e}")

    def send_audio_loop(self):
        """音声データの送信ループ"""
        while True:
            if self.connected and self.audio_handler.state.mic_enabled:
                audio_data = self.audio_handler.get_outgoing_audio()
                if audio_data:
                    try:
                        self.sio.emit('audio_data', {'audio': audio_data.hex()})
                    except Exception as e:
                        logger.error(f"Error sending audio data: {e}")
            time.sleep(0.02)  # 20ms interval

    def reconnect(self):
        """再接続処理"""
        if not self.connected and self.reconnect_attempts < self.max_reconnect_attempts:
            try:
                logger.info(f"Attempting to reconnect... (Attempt {self.reconnect_attempts + 1})")
                self.sio.connect(self.server_url)
                self.reconnect_attempts = 0
            except Exception as e:
                self.reconnect_attempts += 1
                logger.error(f"Reconnection failed: {e}")
                time.sleep(min(5 * self.reconnect_attempts, 30))

    def start(self):
        """クライアントの開始"""
        try:
            self.audio_handler.initialize()
            self.sio.connect(self.server_url)
            
            # オーディオ送信スレッドの開始
            threading.Thread(target=self.send_audio_loop, daemon=True).start()
            
            # メインループ
            while True:
                time.sleep(1)
                if not self.connected:
                    self.reconnect()
        
        except KeyboardInterrupt:
            logger.info("Shutting down...")
        except Exception as e:
            logger.error(f"Client error: {e}")
        finally:
            self.cleanup()

    def cleanup(self):
        """リソースの解放"""
        if self.sio.connected:
            try:
                self.sio.disconnect()
            except Exception as e:
                logger.error(f"Socket disconnect error: {e}")
        self.audio_handler.cleanup()
