#raspi内プログラム
from flask import Flask, request, jsonify
import serial
import time
import requests

app = Flask(__name__)

# Arduinoとの接続設定
SERIAL_PORT = '/dev/ttyACM0'  # Arduinoのポート名
BAUD_RATE = 9600
TIMEOUT = 1

# WindowsのIPアドレスとポート
WINDOWS_CALLBACK_URL = "http://172.25.16.204:9000/angle_feedback"  # WindowsのIPアドレスを適切に設定

try:
   # シリアル接続の初期化
   ser = serial.Serial(SERIAL_PORT, BAUD_RATE, timeout=TIMEOUT)
   time.sleep(2)  # 接続の安定化を待つ
   print(f"[INFO] Arduinoとの接続が確立されました: {SERIAL_PORT}")
except serial.SerialException as e:
   print(f"[ERROR] Arduinoとの接続に失敗しました: {e}")
   ser = None

@app.route('/update_angles', methods=['POST'])
def update_angles():
    if ser is None:
        return jsonify({
            "status": "error",
            "message": "Arduinoとの接続が確立されていません"
        }), 500

    try:
        data = request.get_json()
        servo1 = data.get('servo1', 25)
        servo2 = data.get('servo2', 17)
        servo3 = data.get('servo3', 137)
        servo4 = data.get('servo4', 120)
        
        # Arduino用のフォーマット
        angle_data = f"{servo1},{servo2},{servo3},{servo4}\n"
        
        # Arduinoへデータ送信
        ser.write(angle_data.encode('utf-8'))
        
        # Arduinoからの応答を待機
        time.sleep(0.1)
        arduino_response = None
        if ser.in_waiting > 0:
            arduino_response = ser.readline().decode('utf-8').strip()

        # Windowsにフィードバック送信
        try:
            windows_response = requests.post(
                WINDOWS_CALLBACK_URL,
                json={
                    "servo1": servo1,
                    "servo2": servo2,
                    "servo3": servo3,
                    "servo4": servo4,
                    "arduino_response": arduino_response
                },
                timeout=5
            )
            
            return jsonify({
                "status": "success",
                "message": "角度データを送信しました",
                "angles": {
                    "servo1": servo1,
                    "servo2": servo2,
                    "servo3": servo3,
                    "servo4": servo4
                },
                "arduino_response": arduino_response
            })

        except Exception as e:
            return jsonify({
                "status": "error", 
                "message": str(e)
            }), 500

    except Exception as e:
        return jsonify({
            "status": "error",
            "message": str(e)
        }), 500

@app.route('/health', methods=['GET'])
def health_check():
   """サーバーとArduinoの状態を確認するエンドポイント"""
   arduino_status = "connected" if ser and ser.is_open else "disconnected"
   return jsonify({
       "status": "running",
       "arduino": arduino_status
   })

def cleanup():
   """終了時のクリーンアップ処理"""
   if ser and ser.is_open:
       print("[INFO] シリアル接続を終了します")
       ser.close()

if __name__ == '__main__':
   try:
       app.run(host='0.0.0.0', port=5000)
   finally:
       cleanup()